"""Define representations for :py:class:`ipywidget` comms."""

from __future__ import annotations

import bisect
import logging
import re
from abc import ABCMeta, abstractmethod
from base64 import standard_b64encode
from copy import copy
from datetime import date, datetime
from decimal import Decimal
from functools import partial
from typing import TYPE_CHECKING

from prompt_toolkit.filters.base import Condition
from prompt_toolkit.layout.containers import HSplit, VSplit
from prompt_toolkit.layout.processors import BeforeInput
from prompt_toolkit.widgets.base import Box

from euporie.core.comm.base import Comm, CommView
from euporie.core.data_structures import DiBool
from euporie.core.kernel import MsgCallbacks
from euporie.core.widgets.cell_outputs import CellOutputArea
from euporie.core.widgets.decor import FocusedStyle
from euporie.core.widgets.display import Display
from euporie.core.widgets.forms import (
    Button,
    Checkbox,
    Dropdown,
    Label,
    LabelledWidget,
    Progress,
    Select,
    Slider,
    Swatch,
    Text,
    ToggleButton,
    ToggleButtons,
)
from euporie.core.widgets.layout import AccordionSplit, ReferencedSplit, TabbedSplit

if TYPE_CHECKING:
    from typing import Any, Iterable, MutableSequence, Sequence

    from prompt_toolkit.buffer import Buffer
    from prompt_toolkit.formatted_text.base import AnyFormattedText
    from prompt_toolkit.layout.containers import AnyContainer, _Split

    from euporie.core.comm.base import KernelTab
    from euporie.core.widgets.cell_outputs import OutputParent
    from euporie.core.widgets.forms import SelectableWidget, ToggleableWidget
    from euporie.core.widgets.layout import StackedSplit

    JSONType = str | int | float | bool | None | dict[str, Any] | Iterable[Any]

log = logging.getLogger(__name__)


_binary_types = (memoryview, bytearray, bytes)

WIDGET_MODELS: dict[str, type[IpyWidgetComm]] = {}


def _separate_buffers(
    substate: dict | list | tuple,
    path: list[str | int],
    buffer_paths: list[list[str | int]],
    buffers: MutableSequence[memoryview | bytearray | bytes],
) -> dict | list | tuple:
    """Remove binary types from dicts and lists, but keep track of their paths.

    Any part of the dict/list that needs modification will be cloned, so the original
    stays untouched.

    Args:
        substate: A dictionary or list like containing binary data
        path: A list of dictionary/list keys describing the path root to the substate
        buffer_paths: A list to which binary buffer paths will be added
        buffers: A list to which binary buffers will be added

    Raises:
        ValueError: Raised when the substate is not a list, tuple or dictionary

    Returns:
        The updated substrate without buffers
    """
    cloned_substrate: list | dict | None = None
    if isinstance(substate, (list, tuple)):
        for i, v in enumerate(substate):
            if isinstance(v, _binary_types):
                if cloned_substrate is None:
                    cloned_substrate = list(substate)  # shallow clone list/tuple
                assert isinstance(cloned_substrate, list)
                cloned_substrate[i] = None
                buffers.append(v)
                buffer_paths.append(path + [i])
            elif isinstance(v, (dict, list, tuple)):
                vnew = _separate_buffers(v, path + [i], buffer_paths, buffers)
                if v is not vnew:  # only assign when value changed
                    if cloned_substrate is None:
                        cloned_substrate = list(substate)  # shallow clone list/tuple
                    assert isinstance(cloned_substrate, list)
                    cloned_substrate[i] = vnew
    elif isinstance(substate, dict):
        for k, v in substate.items():
            if isinstance(v, _binary_types):
                if cloned_substrate is None:
                    cloned_substrate = dict(substate)  # shallow clone dict
                del cloned_substrate[k]
                buffers.append(v)
                buffer_paths.append(path + [k])
            elif isinstance(v, (dict, list, tuple)):
                vnew = _separate_buffers(v, path + [k], buffer_paths, buffers)
                if v is not vnew:  # only assign when value changed
                    if cloned_substrate is None:
                        cloned_substrate = dict(substate)  # clone list/tuple
                    cloned_substrate[k] = vnew
    else:
        raise ValueError("expected state to be a list or dict, not %r" % substate)
    return cloned_substrate if cloned_substrate is not None else substate


class IpyWidgetComm(Comm, metaclass=ABCMeta):
    """A Comm object which represents ipython widgets."""

    target_name = "jupyter.widget"

    def __init__(
        self,
        comm_container: KernelTab,
        comm_id: str,
        data: dict,
        buffers: Sequence[bytes],
    ) -> None:
        """Create a new instance of the ipywidget."""
        super().__init__(comm_container, comm_id, data, buffers)
        self.sync = True

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Add ipywidget model classes to a registry when they are created."""
        super().__init_subclass__(**kwargs)
        if cls.__name__.endswith("Model"):
            WIDGET_MODELS[cls.__name__] = cls

    def set_state(self, key: str, value: JSONType) -> None:
        """Send a ``comm_msg`` to the kernel with local state changes."""
        if self.sync:
            self.data.setdefault("state", {})[key] = value
            if self.comm_container.kernel:
                self.comm_container.kernel.kc_comm(
                    comm_id=self.comm_id,
                    data={"method": "update", "state": {key: value}},
                )
            self.update_views({key: value})

    def process_data(self, data: dict, buffers: Sequence[bytes]) -> None:
        """Handle incoming Comm update messages, updating the state and views."""
        # Add buffers to data based on buffer paths
        self.buffers = list(buffers)
        for buffer_path, buffer in zip(
            data.get("buffer_paths", []),
            buffers,
        ):
            parent = data["state"]
            for key in buffer_path[:-1]:
                parent = parent[key]
            parent[buffer_path[-1]] = buffer

        method = data.get("method")
        if method is None:
            self.data = data
        elif method == "update":
            changes = data.get("state", {})
            self.data["state"].update(changes)
            self.update_views(changes)

    @abstractmethod
    def create_view(self, parent: OutputParent) -> CommView:
        """Abstract method for creating a view of the ipywidget."""

    def _get_embed_state(self) -> dict[str, Any]:
        """Convert the ipywidgets state to embeddable json."""
        buffer_paths: list[list[str | int]] = []
        buffers: list[bytes] = []
        state = _separate_buffers(self.data["state"], [], buffer_paths, buffers)
        assert isinstance(state, dict)
        output = {
            "model_name": state.get("_model_name"),
            "model_module": state.get("_model_module"),
            "model_module_version": state.get("_model_module_version"),
            "state": {
                **dict(state.items()),
                "buffers": [
                    {
                        "encoding": "base64",
                        "path": p,
                        "data": standard_b64encode(d).decode("ascii"),
                    }
                    for p, d in zip(buffer_paths, buffers)
                ],
            },
        }
        return output


class UnimplementedModel(IpyWidgetComm):
    """An ipywidget used to represent unimplemented widgets."""

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view."""
        return CommView(Display("[Widget not implemented]", format_="ansi"))


class OutputModel(IpyWidgetComm):
    """An Output ipywidget."""

    model_name = "OutputModel"

    def __init__(
        self,
        comm_container: KernelTab,
        comm_id: str,
        data: dict,
        buffers: Sequence[bytes],
    ) -> None:
        """Create a new output ipywidget instance."""
        super().__init__(comm_container, comm_id, data, buffers)
        self.clear_output_wait = False
        self.prev_msg_id = ""
        self.original_callbacks = MsgCallbacks()
        self.callbacks: MsgCallbacks = MsgCallbacks(
            {
                "add_output": self.add_output,
                "clear_output": self.clear_output,
            }
        )

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of this output ipywidget."""
        container = CellOutputArea(
            self.data.get("state", {}).get("outputs", []), parent
        )
        return CommView(
            container,
            {"outputs": partial(setattr, container, "json")},
        )

    def add_output(self, json: dict[str, Any]) -> None:
        """Add a new output to this widget."""
        if self.clear_output_wait:
            self.set_state("outputs", [json])
        else:
            self.set_state("outputs", [*self.data["state"]["outputs"], json])

    def clear_output(self, wait: bool = False) -> None:
        """Remove all outputs from this widget."""
        if wait:
            self.clear_output_wait = True
        else:
            self.clear_output_wait = False
            self.set_state("outputs", [])

    def process_data(self, data: dict, buffers: Sequence[bytes]) -> None:
        """Modify the callbacks of a given message to add outputs to this ipywidget."""
        if data.get("method") == "update" and self.comm_container.kernel:
            if (msg_id := data.get("state", {}).get("msg_id")) is not None:
                # Replace the kernel callbacks of the given message ID
                if msg_id:
                    # Replace the message's callbacks, saving a copy
                    self.original_callbacks = copy(
                        self.comm_container.kernel.msg_id_callbacks[msg_id]
                    )
                    del self.comm_container.kernel.msg_id_callbacks[msg_id]
                    self.comm_container.kernel.msg_id_callbacks[msg_id].update(
                        self.callbacks
                    )
                else:
                    # Restore the message's callbacks
                    if self.original_callbacks:
                        self.comm_container.kernel.msg_id_callbacks[
                            self.prev_msg_id
                        ] = self.original_callbacks
                        self.original_callbacks = MsgCallbacks()
                    else:
                        del self.comm_container.kernel.msg_id_callbacks[
                            self.prev_msg_id
                        ]
                self.prev_msg_id = msg_id

        super().process_data(data, buffers)


class LayoutIpyWidgetComm(IpyWidgetComm, metaclass=ABCMeta):
    """Base class for layout widgets with children."""

    def render_children(
        self, models: list[str], parent: OutputParent
    ) -> list[AnyContainer]:
        """Create views for the child Comms in the layout."""
        return [
            self.comm_container.comms[
                ipy_model[ipy_model.startswith("IPY_MODEL_") and len("IPY_MODEL_") :]
            ].new_view(parent)
            for ipy_model in models
        ]

    def box_style(self) -> str:
        """Convert the ipywidget box_style to a prompt_toolkit style string."""
        if style := self.data["state"].get("box_style", ""):
            return f"class:{style}"
        return "class:default"


class BoxModel(LayoutIpyWidgetComm):
    """A box layout ipywidget (basically the same a HBox)."""

    padding = 1
    Split: type[_Split] = VSplit

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of the layout ipywidget."""
        container = ReferencedSplit(
            self.Split,
            self.render_children(self.data["state"]["children"], parent),
            padding=self.padding,
            style=self.box_style,
        )

        def set_children(models: list[str]) -> None:
            """Set the children in they layout view when they change."""
            container.children = self.render_children(models, parent)

        return CommView(
            container,
            {"children": set_children},
        )


class HBoxModel(BoxModel):
    """A horizontal layout ipywidget."""


class VBoxModel(BoxModel):
    """A vertical layout ipywidget."""

    Split = HSplit
    padding = 0


class TabModel(LayoutIpyWidgetComm):
    """A tabbed layout ipywidget."""

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of the tabbed ipywidget."""
        children = self.data["state"]["children"]
        container = TabbedSplit(
            children=self.render_children(children, parent),
            titles=list(self.data["state"].get("_titles", {}).values())
            or [""] * len(children),
            active=self.data["state"].get("selected_index"),
            style=self.box_style,
            on_change=self.update_index,
        )

        def set_children(models: list[str]) -> None:
            """Set the children of the tab view when they change."""
            container.children = self.render_children(models, parent)

        def set_titles(new: dict[int, str]) -> None:
            """Set the titles on the tabs when they change."""
            titles = container.titles
            for index, value in new.items():
                titles[int(index)] = value
            container.titles = titles

        return CommView(
            container,
            {
                "children": set_children,
                "_titles": set_titles,
                "selected_index": partial(setattr, container, "active"),
            },
        )

    def update_index(self, container: StackedSplit) -> None:
        """Send a ``comm_message`` updating the selected index when it changes."""
        self.set_state("selected_index", container.active)


class AccordionModel(LayoutIpyWidgetComm):
    """An accoridon layout ipywidget."""

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of the accordion ipywidget."""
        children = self.data["state"]["children"]
        container = AccordionSplit(
            children=self.render_children(children, parent),
            titles=list(self.data["state"].get("_titles", {}).values())
            or [""] * len(children),
            active=self.data["state"].get("selected_index"),
            style=self.box_style,
            on_change=self.update_index,
        )

        def set_children(models: list[str]) -> None:
            """Set the children of the accordion when they change."""
            container.children = self.render_children(models, parent)

        def set_titles(new: dict[int, str]) -> None:
            """Set the titles on the accordion when they change."""
            titles = container.titles
            for index, value in new.items():
                titles[int(index)] = value
            container.titles = titles

        return CommView(
            container,
            {
                "children": set_children,
                "_titles": set_titles,
                "selected_index": partial(setattr, container, "active"),
            },
        )

    def update_index(self, container: StackedSplit) -> None:
        """Send a ``comm_message`` updating the selected index when it changes."""
        self.set_state("selected_index", container.active)


class ButtonModel(IpyWidgetComm):
    """A Button ipywidget."""

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of the button ipywidget."""
        button = Button(
            text=self.text,
            on_click=self.click,
            style=self.button_style,
            disabled=Condition(lambda: self.data["state"].get("disabled", False)),
        )
        return CommView(
            FocusedStyle(button),
            setters={"description": partial(setattr, button, "text")},
        )

    def text(self) -> str:
        """Generate the button text, optionally including an icon if specified."""
        text = self.data["state"].get("description", "")
        if icon := self.data["state"].get("icon", ""):
            from euporie.core.reference import FA_ICONS

            text = f"{FA_ICONS.get(icon, '#')} {text}"
        return text

    def button_style(self) -> str:
        """Convert the ipywidget button_style to a prompt_toolkit style string."""
        if style := self.data["state"].get("button_style", ""):
            return f"class:ipywidget,{style}"
        return "class:ipywidget"

    def click(self, button: Button) -> None:
        """Send a ``comm_msg`` describing a click event."""
        if self.comm_container.kernel:
            self.comm_container.kernel.kc_comm(
                comm_id=self.comm_id,
                data={"method": "custom", "content": {"event": "click"}},
            )


class TextBoxIpyWidgetComm(IpyWidgetComm, metaclass=ABCMeta):
    """A mixin for ipywidgets which use text-box entry."""

    default_rows = 1
    multiline = False

    def validation(self, x: Any) -> bool:
        """Enure the entered text can be normalized."""
        return self.normalize(x) is not None

    def normalize(self, x: Any) -> Any | None:
        """Enure the selected value is permitted and the correct type."""
        return x

    def value(self) -> str:
        """Return the ipywidget's value."""
        return self.data["state"].get("value", "")

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of the text-box ipywidget."""
        text = Text(
            text=self.value(),
            options=lambda: self.data["state"].get("options", []),
            on_text_changed=self.update_value,
            validation=self.validation,
            height=self.data["state"].get("rows") or self.default_rows,
            multiline=self.multiline,
            placeholder=self.data.get("state", {}).get("placeholder"),
            disabled=Condition(lambda: self.data["state"].get("disabled", False)),
            style="class:ipywidget",
        )
        labelled_widget = LabelledWidget(
            body=text,
            label=lambda: self.data.get("state", {}).get("description", ""),
            style="class:ipywidget",
            html=self.data.get("state", {}).get("description_allow_html", False),
        )
        container = FocusedStyle(labelled_widget)
        return CommView(
            container,
            setters={
                "value": lambda x: setattr(text.buffer, "text", str(x)),
                "rows": partial(setattr, text.text_area.window, "height"),
                "placeholder": partial(setattr, text, "placeholder"),
                "description_allow_html": partial(setattr, labelled_widget, "html"),
            },
        )

    def update_value(self, buffer: Buffer) -> None:
        """Set the selected index when the ipywidget's entered value changes."""
        if (value := self.normalize(buffer.text)) is not None:
            self.set_state("value", value)


class TextModel(TextBoxIpyWidgetComm):
    """A text input widget."""


class TextareaModel(TextBoxIpyWidgetComm):
    """A text input widget."""

    default_rows = 3
    multiline = True


class IntOptionsMixin:
    """A mixin for ipywidgets which accept a range of integer values."""

    data: dict[str, Any]

    def normalize(self, x: Any) -> int | None:
        """Enure the selected value is within the permitted range and is a integer."""
        try:
            value = int(x)
        except ValueError:
            return None
        else:
            if minimum := self.data.get("state", {}).get("min"):
                if value < minimum:
                    return None
            if maximum := self.data.get("state", {}).get("max"):
                if maximum < value:
                    return None
            return value

    @property
    def options(self) -> list[int]:
        """Generate a list of available options in a range of integers."""
        step = self.data["state"].get("step", 1)
        return list(
            range(
                self.data["state"].get("min", 0),
                self.data["state"].get("max", 100) + step,
                step,
            )
        )


class FloatOptionsMixin:
    """A mixin for ipywidgets which accept a range of float values."""

    data: dict[str, Any]

    def normalize(self, x: Any) -> float | None:
        """Enure the selected value is within the permitted range and is a float."""
        try:
            value = float(x)
        except ValueError:
            return None
        else:
            if minimum := self.data.get("state", {}).get("min"):
                if value <= minimum:
                    return None
            if maximum := self.data.get("state", {}).get("max"):
                if maximum <= value:
                    return None
            return value

    @property
    def options(self) -> list[float]:
        """Generate a list of available options in a range of floats."""
        step = Decimal(str(self.data["state"].get("step", 0.1)))
        start = Decimal(str(self.data["state"].get("min", 0)))
        stop = Decimal(str(self.data["state"].get("max", 100))) + step
        options = [float(start + step * i) for i in range(int((stop - start) / step))]
        # Ensure value is in list of options
        value = self.data["state"].get("value", 0.0)
        if value not in options:
            if isinstance(value, list):
                values = value
            else:
                values = [value]
            for value in values:
                bisect.insort(options, value)
        return options


class FloatLogOptionsMixin(FloatOptionsMixin):
    """A mixin for ipywidgets which accept a value from range of exponents."""

    data: dict[str, Any]

    def normalize(self, x: Any) -> float | None:
        """Enure the selected value is within the permitted range and is a float."""
        try:
            value = float(x)
        except ValueError:
            return None
        else:
            base = Decimal(str(self.data["state"].get("base", 10)))
            start = Decimal(str(self.data["state"].get("min", 0)))
            if value <= base**start:
                return None
            stop = Decimal(str(self.data["state"].get("max", 4)))
            if base**stop <= value:
                return None
            return value

    @property
    def options(self) -> list[float]:
        """Generate a list of available options in a range of log values."""
        base = Decimal(str(self.data["state"].get("base", 10)))
        start = Decimal(str(self.data["state"].get("min", 0)))
        step = Decimal(str(self.data["state"].get("step", 1)))
        stop = Decimal(str(self.data["state"].get("max"))) + step
        options = [
            float(base ** (start + step * i)) for i in range(int((stop - start) / step))
        ]
        # Ensure value is in list of options
        value = self.data["state"].get("value", 0.0)
        if value not in options:
            bisect.insort(options, value)
        return options


class SliderIpyWidgetComm(IpyWidgetComm, metaclass=ABCMeta):
    """Base class for slider ipywidgets."""

    def normalize(self, x: Any) -> Any:
        """Convert the internal widget's value to one compatible with the ipywidget."""
        return x

    @property
    @abstractmethod
    def options(self) -> list:
        """Abstract method to return a list of available options."""
        return []

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of the slider ipywidget."""
        vertical = Condition(
            lambda: self.data["state"].get("orientation", "horizontal") == "vertical"
        )
        options = self.options
        slider = Slider(
            options=options,
            indices=self.indices,
            multiple=False,
            on_change=self.update_value,
            style="class:ipywidget",
            arrows=(
                lambda: "▼" if vertical() else "⮜",
                lambda: "⮝" if vertical() else "▶",
            ),
            show_arrows=True,
            vertical=Condition(
                lambda: self.data["state"].get("orientation", "horizontal")
                == "vertical"
            ),
            disabled=Condition(lambda: self.data["state"].get("disabled", False)),
        )

        labelled = LabelledWidget(
            body=slider,
            label=lambda: self.data["state"].get("description", ""),
            style="class:ipywidget",
            vertical=Condition(
                lambda: self.data["state"].get("orientation", "horizontal")
                == "vertical"
            ),
        )
        return CommView(
            FocusedStyle(labelled),
            setters={
                "value": partial(self.set_value, slider),
                "index": partial(self.set_value, slider),
                "_options_labels": partial(setattr, slider, "options"),
                "min": partial(lambda x: setattr(slider, "options", self.options)),
                "max": partial(lambda x: setattr(slider, "options", self.options)),
                "step": partial(lambda x: setattr(slider, "options", self.options)),
            },
        )

    @property
    def indices(self) -> list[Any]:
        """Return the selected index as a list."""
        return [
            self.options.index(value) for value in [self.data["state"].get("value", 0)]
        ]

    def update_value(self, container: SelectableWidget) -> None:
        """Send a ``comm_message`` updating the value when it changes."""
        if (index := container.index) is not None:
            if (value := self.normalize(container.options[index])) is not None:
                self.set_state("value", value)

    def set_value(self, slider: Slider, value: Any) -> None:
        """Set the selected index when the ipywidget's selected value changes."""
        if value in slider.options:
            slider.index = slider.options.index(value)
            slider.value_changed()


class RangeSliderIpyWidgetComm(SliderIpyWidgetComm):
    """Base class for range slider ipywidgets."""

    @property
    def indices(self) -> list[int]:
        """Return the first and last selected indices."""
        output = []
        for value in self.data["state"]["value"]:
            if value in self.options:
                output.append(self.options.index(value))
        return output

    def update_value(self, slider: SelectableWidget) -> None:
        """Send a ``comm_message`` updating the values when they change."""
        self.set_state(
            "value", [self.normalize(slider.options[index]) for index in slider.indices]
        )

    def set_value(self, slider: Slider, values: Any) -> None:
        """Any float value is permitted - we might need to add an option."""
        indices = slider.indices
        norm_values = [self.normalize(value) for value in values]
        for i, value in enumerate(norm_values):
            # Ensure value is in list of options
            if value not in slider.options:
                bisect.insort(slider.options, value)
            indices[i] = slider.options.index(value)
        slider.indices = indices
        slider.value_changed()


class IntSliderModel(IntOptionsMixin, SliderIpyWidgetComm):
    """A slider ipywidget that accepts a single integer value."""


class FloatSliderModel(FloatOptionsMixin, SliderIpyWidgetComm):
    """A slider ipywidget that accepts a single float value."""


class FloatLogSliderModel(FloatLogOptionsMixin, SliderIpyWidgetComm):
    """A slider ipywidget that accepts a single value on a log scale."""


class IntRangeSliderModel(IntOptionsMixin, RangeSliderIpyWidgetComm):
    """A slider ipywidget that accepts a range of integer values."""


class FloatRangeSliderModel(FloatOptionsMixin, RangeSliderIpyWidgetComm):
    """A slider ipywidget that accepts a range of float values."""


class NumberTextBoxIpyWidgetComm(TextBoxIpyWidgetComm, metaclass=ABCMeta):
    """Base class for text-box ipywidgets with numerical values."""

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of the numerical text-box ipywidget."""
        disabled = disabled = Condition(
            lambda: self.data["state"].get("disabled", False)
        )

        text = Text(
            text=self.data.get("state", {}).get("value", ""),
            on_text_changed=self.update_value,
            validation=self.validation,
            disabled=disabled,
            style="class:ipywidget",
        )
        container = FocusedStyle(
            LabelledWidget(
                body=ReferencedSplit(
                    VSplit,
                    [
                        text,
                        Button(
                            "-",
                            show_borders=DiBool(True, False, True, True),
                            on_click=self.decr,
                            disabled=disabled,
                            style="class:ipywidget",
                        ),
                        Button(
                            "+",
                            show_borders=DiBool(True, True, True, False),
                            on_click=self.incr,
                            disabled=disabled,
                            style="class:ipywidget",
                        ),
                    ],
                ),
                label=lambda: self.data.get("state", {}).get("description", ""),
                style="class:ipywidget",
            )
        )
        return CommView(
            container,
            setters={"value": partial(lambda x: setattr(text.buffer, "text", str(x)))},
        )

    def incr(self, button: Button) -> None:
        """Increment the widget's value by one step."""
        value = Decimal(str(self.data["state"]["value"]))
        step = Decimal(str(self.data["state"].get("step", 1) or 1))
        if (new := self.normalize(value + step)) is not None:
            self.set_state("value", new)

    def decr(self, button: Button) -> None:
        """Decrement the widget's value by one step."""
        value = Decimal(str(self.data["state"]["value"]))
        step = Decimal(str(self.data["state"].get("step", 1) or 1))
        if (new := self.normalize(value - step)) is not None:
            self.set_state("value", new)


class IntTextModel(IntOptionsMixin, NumberTextBoxIpyWidgetComm):
    """An integer textbox ipwidget."""


class BoundedIntTextModel(IntOptionsMixin, NumberTextBoxIpyWidgetComm):
    """An integer textbox ipwidget with upper and lower bounds."""


class FloatTextModel(FloatOptionsMixin, NumberTextBoxIpyWidgetComm):
    """A float textbox ipwidget."""


class BoundedFloatTextModel(FloatOptionsMixin, NumberTextBoxIpyWidgetComm):
    """An float textbox ipwidget with upper and lower bounds."""


class ProgressIpyWidgetComm(IpyWidgetComm, metaclass=ABCMeta):
    """The base class for progress bar ipywidgets."""

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of the progress bar."""
        vertical = Condition(
            lambda: self.data["state"].get("orientation", "horizontal") == "vertical"
        )
        progress = Progress(
            start=self.data["state"].get("min", 0),
            stop=self.data["state"].get("max", 100),
            step=self.data["state"].get("step", 1),
            value=self.data["state"].get("value", 0),
            vertical=vertical,
            style=self.bar_style,
        )
        container = FocusedStyle(
            LabelledWidget(
                body=progress,
                vertical=vertical,
                label=lambda: self.data.get("state", {}).get("description", ""),
                style="class:ipywidget",
            ),
        )
        return CommView(
            container,
            setters={
                "value": partial(setattr, progress, "value"),
                "min": partial(setattr, progress.control, "start"),
                "max": partial(setattr, progress.control, "stop"),
            },
        )

    def bar_style(self) -> str:
        """Convert the ipywidget ``bar_style`` to a prompt_toolkit style string."""
        if style := self.data["state"].get("bar_style", ""):
            return f"class:{style}"
        return ""


class IntProgressModel(IntOptionsMixin, ProgressIpyWidgetComm):
    """A progress bar ipywidget that accepts integer values."""


class FloatProgressModel(FloatOptionsMixin, ProgressIpyWidgetComm):
    """A progress bar ipywidget that accepts float values."""


class ToggleableIpyWidgetComm(IpyWidgetComm, metaclass=ABCMeta):
    """Base class for toggleable ipywidgets."""

    def normalize(self, x: Any) -> float | None:
        """Cat the container's selected value to a bool if possible."""
        try:
            value = bool(x)
        except ValueError:
            return None
        else:
            return value

    def value_changed(self, button: ToggleableWidget) -> None:
        """Send a ``comm_message`` updating the value when it changes."""
        if (value := self.normalize(button.selected)) is not None:
            self.set_state("value", value)


class ToggleButtonModel(ToggleableIpyWidgetComm):
    """A toggleable button ipywidget."""

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of the toggle button ipywidget."""
        button = ToggleButton(
            text=self.text,
            on_click=self.value_changed,
            selected=self.data["state"].get("value", False),
            style=self.button_style,
            disabled=Condition(lambda: self.data["state"].get("disabled", False)),
        )
        return CommView(
            FocusedStyle(
                button,
            ),
            setters={"value": partial(setattr, button, "selected")},
        )

    def text(self) -> str:
        """Generate the button text, optionally including an icon if specified."""
        text = self.data["state"].get("description", "")
        if icon := self.data["state"].get("icon", ""):
            from euporie.core.reference import FA_ICONS

            text = f"{FA_ICONS.get(icon, '#')} {text}"
        return text

    def button_style(self) -> str:
        """Convert the ipywidget button_style to a prompt_toolkit style string."""
        if style := self.data["state"].get("button_style", ""):
            return f"class:ipywidget,{style}"
        return "class:ipywidget"


class CheckboxModel(ToggleableIpyWidgetComm):
    """A checkbox ipywidget."""

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of the checkbox ipywidget."""
        checkbox = Checkbox(
            text=self.data["state"].get("description", ""),
            on_click=self.value_changed,
            selected=self.data["state"]["value"],
            disabled=Condition(lambda: self.data["state"].get("disabled", False)),
        )
        return CommView(
            FocusedStyle(
                checkbox,
            ),
            setters={"value": partial(setattr, checkbox, "selected")},
        )


class ValidModel(ToggleableIpyWidgetComm):
    """A validity indicator ipywidget."""

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of the validity ipywidget."""
        checkbox = Checkbox(
            on_click=self.value_changed,
            selected=self.data["state"].get("value", False),
            prefix=("❌", "✔️"),
            style="class:valid",
        )
        return CommView(
            FocusedStyle(
                LabelledWidget(
                    checkbox,
                    label=lambda: self.data["state"].get("description", ""),
                    style="class:ipywidget",
                ),
            ),
            setters={"value": partial(setattr, checkbox, "selected")},
        )


class SelectableIpyWidgetComm(IpyWidgetComm, metaclass=ABCMeta):
    """Base class for selectable ipywidgets."""

    def update_index(self, container: SelectableWidget) -> None:
        """Send a ``comm_message`` updating the selected index when it changes."""
        self.set_state("index", container.index)


class DropdownModel(SelectableIpyWidgetComm):
    """An ipywidget allowing an item to be selected using a drop-down menu."""

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of the drop-down widget."""
        dropdown = Dropdown(
            options=self.data["state"].get("_options_labels", []),
            index=self.data["state"]["index"],
            on_change=self.update_index,
            style="class:ipywidget",
            disabled=Condition(lambda: self.data["state"].get("disabled", False)),
        )
        return CommView(
            FocusedStyle(
                LabelledWidget(
                    body=dropdown,
                    label=lambda: self.data["state"].get("description", ""),
                    style="class:ipywidget",
                )
            ),
            setters={
                "index": partial(setattr, dropdown, "index"),
                "_options_labels": partial(setattr, dropdown, "options"),
            },
        )


class RadioButtonsModel(SelectableIpyWidgetComm):
    """An ipywidget allowing an item to be selected using radio buttons."""

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of the radio-buttons widget."""
        select = Select(
            options=self.data["state"].get("_options_labels", []),
            index=self.data["state"]["index"],
            on_change=self.update_index,
            style="class:ipywidget,radio-buttons",
            prefix=("○", "◉"),
            multiple=False,
            border=None,
            disabled=Condition(lambda: self.data["state"].get("disabled", False)),
            rows=None,
        )
        return CommView(
            FocusedStyle(
                LabelledWidget(
                    body=select,
                    label=lambda: self.data["state"].get("description", ""),
                    style="class:ipywidget",
                )
            ),
            setters={
                "index": partial(setattr, select, "index"),
                "_options_labels": partial(setattr, select, "options"),
            },
        )


class SelectModel(SelectableIpyWidgetComm):
    """An ipywidget allowing a value to be selected from a list of options."""

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of the select widget."""
        select = Select(
            options=self.data["state"].get("_options_labels", []),
            index=self.data["state"]["index"],
            on_change=self.update_index,
            style="class:select",
            multiple=False,
            rows=self.data["state"].get("rows", 5),
            disabled=Condition(lambda: self.data["state"].get("disabled", False)),
        )
        return CommView(
            FocusedStyle(
                LabelledWidget(
                    body=select,
                    label=lambda: self.data["state"].get("description", ""),
                    style="class:ipywidget",
                )
            ),
            setters={
                "index": partial(setattr, select, "index"),
                "_options_labels": partial(setattr, select, "options"),
                "rows": partial(setattr, select, "rows"),
            },
        )


class SelectMultipleModel(IpyWidgetComm):
    """An ipywidget allowing one or more value to be selected from a list of options."""

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of the multiple select widget."""
        select = Select(
            options=self.data["state"].get("_options_labels", []),
            indices=self.data["state"]["index"],
            on_change=self.update_index,
            style="class:select",
            multiple=True,
            rows=self.data["state"]["rows"],
            disabled=Condition(lambda: self.data["state"].get("disabled", False)),
        )
        return CommView(
            FocusedStyle(
                LabelledWidget(
                    body=select,
                    label=lambda: self.data["state"].get("description", ""),
                    style="class:ipywidget",
                )
            ),
            setters={
                "index": partial(setattr, select, "indices"),
                "_options_labels": partial(setattr, select, "options"),
                "rows": partial(setattr, select, "rows"),
            },
        )

    def update_index(self, container: SelectableWidget) -> None:
        """Send a ``comm_message`` updating the selected index when it changes."""
        self.set_state("index", container.indices)


class SelectionSliderModel(SliderIpyWidgetComm):
    """A slider ipywidget where one of a list of options can be selected."""

    @property
    def options(self) -> list[str]:
        """Return a list of the available options."""
        return self.data["state"].get("_options_labels", [])

    @property
    def indices(self) -> list[int]:
        """Return the selected index as a list."""
        return [self.data["state"]["index"]]

    def update_value(self, slider: SelectableWidget) -> None:
        """Send a ``comm_message`` updating the selected index when it changes."""
        self.set_state("index", slider.index)

    def set_value(self, slider: Slider, index: int) -> None:
        """Set the selected index on the selection slider when the value changes."""
        self.sync = False
        slider.control.set_index(ab=index)
        self.sync = True


class SelectionRangeSliderModel(SliderIpyWidgetComm):
    """A slider ipywidget where one or more of a list of options can be selected."""

    @property
    def options(self) -> list[str]:
        """Return a list of the available options."""
        return self.data["state"].get("_options_labels", [])

    @property
    def indices(self) -> list[int]:
        """Return a list of the selected indices."""
        return self.data["state"]["index"]

    def update_value(self, slider: SelectableWidget) -> None:
        """Send a ``comm_message`` updating the selected indices when they change."""
        self.set_state("index", slider.indices)

    def set_value(self, slider: Slider, indices: tuple[int, int]) -> None:
        """Set the selected indices on the selection slider when the value changes."""
        self.sync = False
        for i, index in enumerate(indices):
            slider.control.set_index(handle=i, ab=index)
        self.sync = True


class ToggleButtonsModel(IpyWidgetComm):
    """An ipywidget where a single value can be selected using togglable buttons."""

    def get_label(self, index: int) -> AnyFormattedText:
        """Return the label for each toggle button (optionally including the icon)."""
        label = str(self.data["state"].get("_options_labels", [])[index])
        if index < len(self.data["state"].get("icons", [])):
            if icon := self.data["state"]["icons"][index]:
                from euporie.core.reference import FA_ICONS

                label = f"{FA_ICONS.get(icon, '#')} {label}"
        return label

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of the toggle button widget."""
        options = self.data["state"].get("_options_labels", [])
        buttons = ToggleButtons(
            options=options,
            labels=[partial(self.get_label, i) for i in range(len(options))],
            index=self.data["state"]["index"],
            on_change=self.update_index,
            style=self.button_style,
            multiple=False,
            disabled=Condition(lambda: self.data["state"].get("disabled", False)),
        )
        return CommView(
            FocusedStyle(
                LabelledWidget(
                    buttons,
                    label=lambda: self.data["state"].get("description", ""),
                    style="class:ipywidget",
                )
            ),
            setters={
                "_options_labels": lambda x: self.set_options(buttons),
                "index": partial(setattr, buttons, "index"),
            },
        )

    def set_options(self, buttons: ToggleButtons) -> None:
        """Set the list of selectable options in the toggle buttons."""
        options = self.data["state"].get("_options_labels", [])
        buttons.options = options
        # Update the button label functions in case the number of options has changed
        buttons.labels = [partial(self.get_label, i) for i in range(len(options))]

    def button_style(self) -> str:
        """Convert the ipywidget button_style to a prompt_toolkit style string."""
        if style := self.data["state"].get("button_style", ""):
            return f"class:ipywidget,{style}"
        return "class:ipywidget"

    def update_index(self, container: SelectableWidget) -> None:
        """Set the selected index be sending a comm update message."""
        self.set_state("index", container.index)


class ComboboxModel(TextBoxIpyWidgetComm):
    """A combobox input widget."""

    def normalize(self, x: str) -> str | None:
        """Enure that the entered text matches a permitted option, if required."""
        if self.data["state"].get("ensure_option", False):
            if x in self.data["state"].get("options", []):
                return x
            return None
        else:
            return x


class LabelModel(IpyWidgetComm):
    """A label ipwidget."""

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of the Label widget."""
        label = Label(lambda: self.data["state"].get("value", ""))
        return CommView(label)


class HTMLModel(IpyWidgetComm):
    """A label ipywidget which displays HTML."""

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of the HTML widget."""
        html = Display(
            data=self.data["state"].get("value", ""),
            format_="html",
        )
        return CommView(
            Box(
                html,
                padding_left=0,
                padding_right=0,
            ),
            setters={
                "value": partial(setattr, html, "data"),
            },
        )


class HTMLMathModel(HTMLModel):
    """Alia for :class:`HTMLModel`, which can render maths."""


class ImageModel(IpyWidgetComm):
    """A ipywidget which displays an image."""

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of the image widget."""
        display = Display(
            data=self.data["state"].get("value", b""),
            format_="png",
            px=int(self.data["state"].get("width", 0)),
            py=int(self.data["state"].get("height", 0)),
        )
        box = Box(
            display,
            padding_left=0,
            padding_right=0,
        )

        return CommView(
            box,
            setters={
                "value": partial(setattr, display, "data"),
                "width": lambda x: partial(setattr, display, "px")(int(x)),
                "height": lambda x: partial(setattr, display, "py")(int(x)),
            },
        )


class DatePickerModel(TextBoxIpyWidgetComm):
    """A date-picker ipywidget."""

    def normalize(self, x: str) -> dict[str, int] | None:
        """Attempt to convert entered text to the internal date representation."""
        if not x:
            return None
        try:
            value = datetime.strptime(x, "%Y-%m-%d").date()
        except ValueError:
            return None
        else:
            return {
                "year": value.year,
                "month": value.month - 1,
                "date": value.day,
            }

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of the date-picker widget."""
        comm_view = super().create_view(parent)
        # Wrap the standard value setter to parse the date objects
        value_cb = comm_view.setters["value"]
        comm_view.setters["value"] = partial(
            lambda x: value_cb(self.parse_date(x).strftime("%Y-%m-%d"))
        )
        return comm_view

    def value(self) -> str:
        """Return the text to display in the widget's text area."""
        value = self.data["state"].get("value")
        if value:
            return self.parse_date(value).strftime("%Y-%m-%d")
        else:
            return ""

    def parse_date(self, value: dict[str, int]) -> date:
        """Convert the internal date representation to a python date."""
        return date(value["year"], value["month"] + 1, value["date"])


class ColorPickerModel(TextBoxIpyWidgetComm):
    """A color picker ipywidget."""

    _hex_pattern = re.compile(
        r"#(?:[a-fA-F0-9]{3}(?:[a-fA-F0-9]{3})?|[a-fA-F0-9]{4}(?:[a-fA-F0-9]{4})?)$"
    )

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new view of the color-picker widget."""
        text = Text(
            text=self.value(),
            options=lambda: self.data["state"].get("options", []),
            on_text_changed=self.update_value,
            validation=self.validation,
            height=self.data["state"].get("rows") or self.default_rows,
            multiline=self.multiline,
            placeholder=self.data.get("state", {}).get("placeholder"),
            show_borders=DiBool(True, True, True, False),
            input_processors=[BeforeInput(" ")],
            disabled=Condition(lambda: self.data["state"].get("disabled", False)),
            style="class:ipywidget",
        )

        container = FocusedStyle(
            LabelledWidget(
                body=VSplit(
                    [
                        Swatch(
                            self.format_color,
                            show_borders=DiBool(True, False, True, True),
                        ),
                        text,
                    ],
                ),
                label=lambda: self.data.get("state", {}).get("description", ""),
                style="class:ipywidget",
            )
        )
        return CommView(
            container,
            setters={
                "value": lambda x: setattr(text.buffer, "text", str(x)),
                "rows": partial(setattr, text.text_area.window, "height"),
                "placeholder": partial(setattr, text, "placeholder"),
            },
        )

    def format_color(self) -> str:
        """Format a color as a hex code for display."""
        from euporie.core.reference import NAMED_COLORS

        # TODO - blend alpha colors with the terminal background
        value = self.value()
        if value in NAMED_COLORS:
            value = NAMED_COLORS[value]
        elif 4 <= len(value) < 7:
            value = f"#{value[1]*2}{value[2]*2}{value[3]*2}"
        else:
            value = value[:7]
        return value

    def normalize(self, x: str) -> str | None:
        """Return the color string if it is recognised as an allowed color."""
        from euporie.core.reference import NAMED_COLORS

        if x in NAMED_COLORS or self._hex_pattern.match(x) is not None:
            return x
        else:
            return None


def open_comm_ipywidgets(
    comm_container: KernelTab,
    comm_id: str,
    data: dict,
    buffers: Sequence[bytes],
) -> IpyWidgetComm:
    """Create a new Comm for an :py:mod:`ipywidgets` widget.

    The relevant widget model is selected based on the model name given in the
    ``comm_open`` message data.

    Args:
        comm_container: The notebook this Comm belongs to
        comm_id: The ID of the Comm
        data: The data field from the ``comm_open`` message
        buffers: The buffers field from the ``comm_open`` message

    Returns:
        The initialized widget Comm object.
    """
    model_name = data.get("state", {}).get("_model_name")
    # Skip type checking due to https://github.com/python/mypy/issues/3115
    return WIDGET_MODELS.get(model_name, UnimplementedModel)(
        comm_container, comm_id, data, buffers
    )  # type: ignore



"""Contain dpdated ANSI parsing and Formatted Text processing."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from prompt_toolkit.filters import to_filter
from prompt_toolkit.formatted_text import (
    fragment_list_to_text,
    split_lines,
    to_formatted_text,
)
from prompt_toolkit.layout.margins import ConditionalMargin
from prompt_toolkit.layout.processors import DynamicProcessor, Processor, Transformation
from prompt_toolkit.widgets import TextArea

from euporie.core.margins import NumberedDiffMargin

if TYPE_CHECKING:
    from typing import Any

    from prompt_toolkit.filters import FilterOrBool
    from prompt_toolkit.formatted_text import AnyFormattedText, StyleAndTextTuples
    from prompt_toolkit.layout.processors import TransformationInput

log = logging.getLogger(__name__)


class FormattedTextProcessor(Processor):
    """Apply formatted text to a TextArea."""

    def __init__(self, formatted_text: StyleAndTextTuples) -> None:
        """Initiate the processor.

        Args:
            formatted_text: The text in a buffer but with formatting applied.

        """
        self.formatted_lines: list[StyleAndTextTuples] = []
        self.formatted_text = formatted_text
        super().__init__()

    def apply_transformation(
        self, transformation_input: TransformationInput
    ) -> Transformation:
        """Apply text formatting to a line in a buffer."""
        if not self.formatted_lines:
            self.formatted_lines = list(split_lines(self.formatted_text))
        lineno = transformation_input.lineno
        max_lineno = len(self.formatted_lines) - 1
        if lineno > max_lineno:
            lineno = max_lineno
        line = self.formatted_lines[lineno]
        return Transformation(line)


class FormattedTextArea(TextArea):
    """Apply formatted text to a TextArea."""

    _formatted_text: AnyFormattedText

    def _set_formatted_text(self, value: AnyFormattedText) -> None:
        self._formatted_text = value
        self.text = fragment_list_to_text(self.formatted_text)

    def __init__(
        self,
        formatted_text: AnyFormattedText,
        *args: Any,
        line_numbers: FilterOrBool = False,
        **kwargs: Any,
    ) -> None:
        """Initialize a `FormattedTextArea` instance.

        Args:
            formatted_text: A list of `(style, text)` tuples to display.
            line_numbers: Determines if line numbers are shown,
            args: Arguments to pass to `prompt_toolkit.widgets.TextArea`.
            kwargs: Key-word arguments to pass to `prompt_toolkit.widgets.TextArea`.

        """
        self._formatted_text = formatted_text
        self.line_numbers = to_filter(line_numbers)
        input_processors = kwargs.pop("input_processors", [])
        input_processors.append(DynamicProcessor(self.get_processor))
        # The following is not type checked due to a currently open mypy bug
        # https://github.com/python/mypy/issues/6799
        super().__init__(
            *args,
            input_processors=input_processors,
            **kwargs,
        )  # type: ignore
        # Set the left margins
        self.window.left_margins = [
            ConditionalMargin(
                NumberedDiffMargin(),
                self.line_numbers,
            )
        ]
        # Set the formatted text to display
        self._set_formatted_text(formatted_text)

    @property
    def formatted_text(self) -> StyleAndTextTuples:
        """The formatted text."""
        ft = to_formatted_text(self._formatted_text)
        text = fragment_list_to_text(ft)
        if self.text != text:
            self.text = text
        return ft

    @formatted_text.setter
    def formatted_text(self, value: AnyFormattedText) -> None:
        """Set the formatted text."""
        self._set_formatted_text(value)

    def get_processor(self) -> FormattedTextProcessor:
        """Generate a processor for the formatted text."""
        return FormattedTextProcessor(self.formatted_text)



"""Contain function which convert data to rich format."""

from __future__ import annotations

from typing import TYPE_CHECKING

from euporie.core.convert.core import register
from euporie.core.convert.utils import have_modules

if TYPE_CHECKING:
    from pathlib import Path

    from rich.markdown import Markdown


@register(
    from_="markdown",
    to="rich",
    filter_=have_modules("rich"),
)
async def markdown_to_rich_py(
    data: str | bytes,
    width: int | None = None,
    height: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    path: Path | None = None,
    initial_format: str = "",
) -> Markdown:
    """Convert base64 encoded data to bytes."""
    from rich.markdown import Markdown

    markup = data.decode() if isinstance(data, bytes) else data
    return Markdown(
        markup,
        # code_theme=str(config.syntax_theme),
        # inline_code_theme=str(config.syntax_theme),
    )



"""Main entry point into euporie.core."""


def main(name: "str" = "launch") -> "None":
    """Load and launches the application."""
    from importlib.metadata import entry_points

    for entry in entry_points().get("euporie.apps", []):
        if entry.name == name:
            return entry.load().launch()
    else:
        raise Exception(f"Euporie app `{name}` not installed")


if __name__ == "__main__":
    main()



"""Define custom controls which re-render on resize."""

from __future__ import annotations

import logging
import weakref
from abc import ABCMeta
from functools import partial
from math import ceil
from typing import TYPE_CHECKING

from prompt_toolkit.cache import SimpleCache
from prompt_toolkit.data_structures import Point
from prompt_toolkit.filters.app import has_completions
from prompt_toolkit.filters.base import Condition
from prompt_toolkit.filters.utils import to_filter
from prompt_toolkit.formatted_text.base import to_formatted_text
from prompt_toolkit.formatted_text.utils import fragment_list_width, split_lines
from prompt_toolkit.layout.containers import ConditionalContainer, Float, VSplit, Window
from prompt_toolkit.layout.controls import GetLinePrefixCallable, UIContent, UIControl
from prompt_toolkit.layout.mouse_handlers import MouseHandlers
from prompt_toolkit.layout.screen import Char, WritePosition
from prompt_toolkit.mouse_events import MouseEvent, MouseEventType
from prompt_toolkit.utils import Event, to_str

from euporie.core.commands import add_cmd
from euporie.core.convert.core import convert, find_route
from euporie.core.convert.utils import data_pixel_size, pixels_to_cell_size
from euporie.core.current import get_app
from euporie.core.data_structures import DiInt
from euporie.core.filters import (
    display_has_focus,
    has_dialog,
    has_menus,
    in_tmux,
    scrollable,
)
from euporie.core.ft.utils import wrap
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)
from euporie.core.margins import MarginContainer, ScrollbarMargin
from euporie.core.terminal import tmuxify
from euporie.core.widgets.page import BoundedWritePosition

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any, Callable, Iterable

    from prompt_toolkit.filters import FilterOrBool
    from prompt_toolkit.formatted_text import StyleAndTextTuples
    from prompt_toolkit.key_binding import KeyBindingsBase
    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.layout.containers import AnyContainer
    from prompt_toolkit.layout.dimension import AnyDimension
    from prompt_toolkit.layout.screen import Screen


log = logging.getLogger(__name__)


class DisplayControl(UIControl):
    """A data formatter, which displays cell output data.

    It will attempt to display the data in the best way possible, and reacts to resize
    events - i.e. images are downscaled to fit, markdown is re-flowed, etc.
    """

    def __init__(
        self,
        data: Any,
        format_: str,
        path: Path | None = None,
        fg_color: str | None = None,
        bg_color: str | None = None,
        sizing_func: Callable[[], tuple[int, float]] | None = None,
        focusable: FilterOrBool = False,
        focus_on_click: FilterOrBool = False,
        wrap_lines: FilterOrBool = False,
    ) -> None:
        """Create a new data formatter control.

        Args:
            data: Raw cell output data
            format_: The conversion format of the data to render
            path: The path to the data's original location
            fg_color: The foreground colour to use when rendering this output
            bg_color: The background colour to use when rendering this output
            sizing_func: Function which returns the maximum width and aspect ratio of
                the output
            focusable: Whether the control can be focused
            focus_on_click: Whether to focus the control when clicked
            wrap_lines: Whether lines of output should be wrapped

        """
        self._data = data
        self.format_ = format_
        self.path = path
        self.fg_color = fg_color
        self.bg_color = bg_color
        self.focusable = to_filter(focusable)
        self.focus_on_click = to_filter(focus_on_click)
        self.wrap_lines = to_filter(wrap_lines)
        self.key_bindings = load_registered_bindings(
            "euporie.core.widgets.display.Display"
        )
        self._cursor_position = Point(x=0, y=0)
        self.dy = 0
        self.app = get_app()

        # Whenever the data changes, the UI has to be updated.
        self.on_data_changed = Event(self)
        # Whenever the cursor position changes, the UI has to be updated.
        self.on_cursor_position_changed = Event(self)
        # Allow invalidation events to be added to a specific instance of this control
        self.invalidate_events: list[Event] = [
            self.on_data_changed,
            self.on_cursor_position_changed,
        ]

        self.sizing_func = sizing_func or (lambda: (0, 0))
        self._max_cols = 0
        self._aspect = 0.0

        self.rendered_lines: list[StyleAndTextTuples] = []
        self._format_cache: SimpleCache = SimpleCache(maxsize=50)
        self._content_cache: SimpleCache = SimpleCache(maxsize=50)
        self._size_cache: SimpleCache = SimpleCache(maxsize=1)

    def reset(self) -> None:
        """Clear the display control's caches (required if the control's data changes)."""
        self._format_cache.clear()
        self._size_cache.clear()
        self._content_cache.clear()

    @property
    def data(self) -> Any:
        """Return the control's display data."""
        return self._data

    @data.setter
    def data(self, value: Any) -> None:
        """Set the control's data."""
        self._data = value
        self.reset()
        self.on_data_changed.fire()

    def get_key_bindings(self) -> KeyBindingsBase | None:
        """Return the control's key bindings."""
        return self.key_bindings

    @property
    def cursor_position(self) -> Point:
        """Get the cursor position."""
        return self._cursor_position

    @cursor_position.setter
    def cursor_position(self, value: Point) -> None:
        """Set the cursor position."""
        changed = self._cursor_position != value
        self._cursor_position = value
        if changed:
            self.on_cursor_position_changed.fire()

    def move_cursor_down(self) -> None:
        """Move the cursor down one line."""
        x, y = self.cursor_position
        self.cursor_position = Point(x=x, y=y + 1)

    def move_cursor_up(self) -> None:
        """Move the cursor up one line."""
        x, y = self.cursor_position
        self.cursor_position = Point(x=x, y=max(0, y - 1))

    def move_cursor_left(self) -> None:
        """Move the cursor down one line."""
        x, y = self.cursor_position
        self.cursor_position = Point(x=max(0, x - 1), y=y)

    def move_cursor_right(self) -> None:
        """Move the cursor up one line."""
        x, y = self.cursor_position
        self.cursor_position = Point(x=x + 1, y=y)

    def is_focusable(self) -> bool:
        """Determine if the current control is focusable."""
        return self.focusable()

    def size(self) -> None:
        """Load the maximum cell width and aspect ratio of the output."""
        self._max_cols, self._aspect = self._size_cache.get(
            self.app.render_counter, self.sizing_func
        )

    def hide(self) -> None:
        """Hide the output from show."""
        pass

    @property
    def max_cols(self) -> int:
        """Load the maximum width of the output in terminal columns."""
        self.size()
        return self._max_cols

    @property
    def aspect(self) -> float:
        """Lazily load the aspect ratio of the output."""
        self.size()
        return self._aspect

    def preferred_width(self, max_available_width: int) -> int | None:
        """Return the width of the rendered content."""
        self.max_available_width = max_available_width
        return (
            min(self.max_cols, max_available_width)
            if self.max_cols
            else max_available_width
        )

    def get_rendered_lines(
        self, width: int, height: int, wrap_lines: bool = False
    ) -> list[StyleAndTextTuples]:
        """Render the output data."""
        return []

    def preferred_height(
        self,
        width: int,
        max_available_height: int,
        wrap_lines: bool,
        get_line_prefix: GetLinePrefixCallable | None,
    ) -> int:
        """Return the number of lines in the rendered content."""
        if self.aspect:
            return ceil(min(width, self.max_cols) * self.aspect)
        else:
            self.rendered_lines = self.get_rendered_lines(
                width, max_available_height, self.wrap_lines()
            )
            return len(self.rendered_lines)

    def create_content(self, width: int, height: int) -> UIContent:
        """Generate rendered output at a given size.

        Args:
            width: The desired output width
            height: The desired output height

        Returns:
            `UIContent` for the given output size.

        """
        cols = min(self.max_cols, width) if self.max_cols else width
        rows = ceil(cols * self.aspect) if self.aspect else height
        wrap_lines = self.wrap_lines()

        def get_content() -> dict[str, Any]:
            rendered_lines = self.get_rendered_lines(
                width=cols, height=rows, wrap_lines=wrap_lines
            )
            self.rendered_lines = rendered_lines[:]
            line_count = len(rendered_lines)

            def get_line(i: int) -> StyleAndTextTuples:
                # Return blank lines if the renderer expects more content than we have
                line = rendered_lines[i] if i < line_count else []
                # Add space at end of empty lines as that is a possible cursor position
                # This is what PTK does, and fixes a bug where scrolling would stop
                # working when the cursor was on an empty line.
                if not line:
                    line += [("", " ")]
                return line

            return {
                "get_line": get_line,
                "line_count": line_count,
                "menu_position": Point(0, 0),
            }

        # Re-render if the image width changes, or the terminal character size changes
        key = (cols, self.app.term_info.cell_size_px, wrap_lines)
        return UIContent(
            cursor_position=self.cursor_position,
            **self._content_cache.get(key, get_content),
        )

    def mouse_handler(self, mouse_event: MouseEvent) -> NotImplementedOrNone:
        """Mouse handler for this control."""
        if self.focus_on_click() and mouse_event.event_type == MouseEventType.MOUSE_UP:
            self.app.layout.current_control = self
            return None
        return NotImplemented

    def get_invalidate_events(self) -> Iterable[Event[object]]:
        """Return the Window invalidate events."""
        yield from self.invalidate_events

    @property
    def content_width(self) -> int:
        """Return the maximum line length of the content."""
        return max(fragment_list_width(line) for line in self.rendered_lines)

    def close(self) -> None:
        """Remove the displayed object entirely."""
        if not self.app.leave_graphics():
            self.hide()


class DisplayWindow(Window):
    """A window sub-class which can scroll left and right."""

    content: DisplayControl
    vertical_scroll: int

    def _write_to_screen_at_index(
        self,
        screen: Screen,
        mouse_handlers: MouseHandlers,
        write_position: WritePosition,
        parent_style: str,
        erase_bg: bool,
    ) -> None:
        """Enure the :attr:`horizontal_scroll` is recorded."""
        super()._write_to_screen_at_index(
            screen,
            mouse_handlers,
            write_position,
            parent_style,
            erase_bg,
        )
        # Set the horizontal scroll offset on the render info
        # TODO - fix this upstream
        if self.render_info is not None:
            setattr(  # noqa B010
                self.render_info, "horizontal_scroll", self.horizontal_scroll
            )

    def _scroll_right(self) -> None:
        """Scroll window right."""
        info = self.render_info
        if info is None:
            return
        content_width = self.content.content_width
        if self.horizontal_scroll < content_width - info.window_width:
            if info.cursor_position.y <= info.configured_scroll_offsets.right:
                self.content.move_cursor_right()
            self.horizontal_scroll += 1

    def _scroll_left(self) -> None:
        """Scroll window left."""
        info = self.render_info
        if info is None:
            return
        horizontal_scroll = getattr(
            self.render_info, "horizontal_scroll", 0
        )  # noqa B009
        if horizontal_scroll > 0:
            self.content.move_cursor_left()
            self.horizontal_scroll -= 1


class FormattedTextDisplayControl(DisplayControl):
    """A data formatter, which displays cell output data.

    It will attempt to display the data in the best way possible, and reacts to resize
    events - i.e. images are downscaled to fit, markdown is re-flowed, etc.
    """

    def get_rendered_lines(
        self, width: int, height: int, wrap_lines: bool = False
    ) -> list[StyleAndTextTuples]:
        """Get rendered lines from the cache, or generate them."""

        def render_lines() -> list[StyleAndTextTuples]:
            """Render the lines to display in the control."""
            lines = list(
                split_lines(
                    to_formatted_text(
                        convert(
                            data=self.data,
                            from_=self.format_,
                            to="ft",
                            cols=width,
                            rows=height,
                            fg=self.fg_color,
                            bg=self.bg_color,
                            path=self.path,
                        )
                    )
                )
            )
            if wrap_lines:
                lines = [
                    wrapped_line
                    for line in lines
                    for wrapped_line in split_lines(
                        wrap(line, width, truncate_long_words=False)
                    )
                ]

            # Ensure we have enough lines to fill the data's calculated height
            rows = ceil(min(width, self.max_cols) * self.aspect)
            lines.extend([[]] * max(0, rows - len(lines)))

            return lines

        # Re-render if the image width changes, or the terminal character size changes
        key = (width, self.app.term_info.cell_size_px, wrap_lines)
        return self._format_cache.get(key, render_lines)


class GraphicControl(DisplayControl, metaclass=ABCMeta):
    """A base-class for display controls which render terminal graphics."""

    def __init__(
        self,
        data: Any,
        format_: str,
        path: Path | None = None,
        fg_color: str | None = None,
        bg_color: str | None = None,
        sizing_func: Callable[[], tuple[int, float]] | None = None,
        focusable: FilterOrBool = False,
        focus_on_click: FilterOrBool = False,
        scale: float = 0,
        bbox: DiInt | None = None,
    ) -> None:
        """Initialize the graphic control."""
        super().__init__(
            data,
            format_,
            path,
            fg_color,
            bg_color,
            sizing_func,
            focusable,
            focus_on_click,
        )
        self.bbox = bbox or DiInt(0, 0, 0, 0)

        # Record the original pixel size of the imge
        px, py = data_pixel_size(data, format_, fg=fg_color, bg=bg_color)
        if not px or not py:
            log.warning("Cannot determine image size")
            px, py = px or 100, py or 100
        self.px, self.py = px, py

    def create_content(self, width: int, height: int) -> UIContent:
        """Generate rendered output at a given size.

        Args:
            width: The desired output width
            height: The desired output height

        Returns:
            `UIContent` for the given output size.

        """
        cols = min(self.max_cols, width) if self.max_cols else width
        rows = (
            ceil(cols * self.aspect) - self.bbox.top - self.bbox.bottom
            if self.aspect
            else height
        )

        def get_content() -> dict[str, Any]:
            rendered_lines = self.get_rendered_lines(width=cols, height=rows)
            self.rendered_lines = rendered_lines[:]
            line_count = len(rendered_lines)

            def get_line(i: int) -> StyleAndTextTuples:
                # Return blank lines if the renderer expects more content than we have
                line = []
                if i < line_count:
                    line += rendered_lines[i]
                # Add a space at the end, because that is a possible cursor position.
                # This is what PTK does, and fixes a nasty bug which took me ages to
                # track down the source of, where scrolling would stop working when the
                # cursor was on an empty line.
                line += [("", " ")]
                return line

            return {
                "get_line": get_line,
                "line_count": line_count,
                "menu_position": Point(0, 0),
            }

        # Re-render if the image width changes, or the terminal character size changes
        key = (cols, rows, self.app.term_info.cell_size_px, self.bbox)
        return UIContent(
            cursor_position=self.cursor_position,
            **self._content_cache.get(key, get_content),
        )


class SixelGraphicControl(GraphicControl):
    """A graphic control which displays images as sixels."""

    def get_rendered_lines(
        self, width: int, height: int, wrap_lines: bool = False
    ) -> list[StyleAndTextTuples]:
        """Get rendered lines from the cache, or generate them."""
        cell_size_x, cell_size_y = self.app.term_info.cell_size_px

        def render_lines() -> list[StyleAndTextTuples]:
            """Render the lines to display in the control."""
            full_width = width + self.bbox.left + self.bbox.right
            full_height = height + self.bbox.top + self.bbox.bottom
            cmd = str(
                convert(
                    data=self.data,
                    from_=self.format_,
                    to="sixel",
                    cols=full_width,
                    rows=full_height,
                    fg=self.fg_color,
                    bg=self.bg_color,
                    path=self.path,
                )
            )
            if any(self.bbox):
                from sixelcrop import sixelcrop

                cmd = sixelcrop(
                    data=cmd,
                    # Horizontal pixel offset of the displayed image region
                    x=self.bbox.left * cell_size_x,
                    # Vertical pixel offset of the displayed image region
                    y=self.bbox.top * cell_size_y,
                    # Pixel width of the displayed image region
                    w=width * cell_size_x,
                    # Pixel height of the displayed image region
                    h=height * cell_size_y,
                )

            if get_app().config.tmux_graphics:
                cmd = tmuxify(cmd)

            return list(
                split_lines(
                    to_formatted_text(
                        [
                            # Move cursor down and across by image height and width
                            ("", "\n".join((height) * [" " * (width)])),
                            # Save position, then move back
                            ("[ZeroWidthEscape]", "\x1b[s"),
                            # Move cursor up if there is more than one line to display
                            *(
                                [("[ZeroWidthEscape]", f"\x1b[{height-1}A")]
                                if height > 1
                                else []
                            ),
                            ("[ZeroWidthEscape]", f"\x1b[{width}D"),
                            # Place the image without moving cursor
                            ("[ZeroWidthEscape]", cmd),
                            # Restore the last known cursor position (at the bottom)
                            ("[ZeroWidthEscape]", "\x1b[u"),
                        ]
                    )
                )
            )

        key = (width, self.bbox, (cell_size_x, cell_size_y))
        return self._format_cache.get(key, render_lines)


class ItermGraphicControl(GraphicControl):
    """A graphic control which displays images using iTerm's graphics protocol."""

    def convert_data(self, rows: int, cols: int) -> str:
        """Convert the graphic's data to base64 data."""
        data = self.data
        format_ = self.format_

        full_width = cols + self.bbox.left + self.bbox.right
        full_height = rows + self.bbox.top + self.bbox.bottom

        # Crop image if necessary
        if any(self.bbox):
            import io

            image = convert(
                data=data,
                from_=format_,
                to="pil",
                cols=full_width,
                rows=full_height,
                fg=self.fg_color,
                bg=self.bg_color,
                path=self.path,
            )
            if image is not None:
                cell_size_x, cell_size_y = self.app.term_info.cell_size_px
                # Downscale image to fit target region for precise cropping
                image.thumbnail((full_width * cell_size_x, full_height * cell_size_y))
                image = image.crop(
                    (
                        self.bbox.left * cell_size_x,  # left
                        self.bbox.top * cell_size_y,  # top
                        (self.bbox.left + cols) * cell_size_x,  # right
                        (self.bbox.top + rows) * cell_size_y,  # bottom
                    )
                )
                with io.BytesIO() as output:
                    image.save(output, format="PNG")
                    data = output.getvalue()
                format_ = "png"

        if format_.startswith("base64-"):
            b64data = data
        else:
            b64data = convert(
                data=data,
                from_=format_,
                to="base64-png",
                cols=full_width,
                rows=full_height,
                fg=self.fg_color,
                bg=self.bg_color,
                path=self.path,
            )
        b64data = b64data.replace("\n", "").strip()
        return b64data

    def get_rendered_lines(
        self, width: int, height: int, wrap_lines: bool = False
    ) -> list[StyleAndTextTuples]:
        """Get rendered lines from the cache, or generate them."""

        def render_lines() -> list[StyleAndTextTuples]:
            """Render the lines to display in the control."""
            b64data = self.convert_data(cols=width, rows=height)
            cmd = f"\x1b]1337;File=inline=1;width={width}:{b64data}\a"
            return list(
                split_lines(
                    to_formatted_text(
                        [
                            # Move cursor down and across by image height and width
                            ("", "\n".join((height) * [" " * (width)])),
                            # Save position, then move back
                            ("[ZeroWidthEscape]", "\x1b[s"),
                            # Move cursor up if there is more than one line to display
                            *(
                                [("[ZeroWidthEscape]", f"\x1b[{height-1}A")]
                                if height > 1
                                else []
                            ),
                            ("[ZeroWidthEscape]", f"\x1b[{width}D"),
                            # Place the image without moving cursor
                            ("[ZeroWidthEscape]", tmuxify(cmd)),
                            # Restore the last known cursor position (at the bottom)
                            ("[ZeroWidthEscape]", "\x1b[u"),
                        ]
                    )
                )
            )

        key = (width, self.bbox, self.app.term_info.cell_size_px)
        return self._format_cache.get(key, render_lines)


_kitty_image_count = 1


class KittyGraphicControl(GraphicControl):
    """A graphic control which displays images using Kitty's graphics protocol."""

    def __init__(
        self,
        data: Any,
        format_: str,
        path: Path | None = None,
        fg_color: str | None = None,
        bg_color: str | None = None,
        sizing_func: Callable[[], tuple[int, float]] | None = None,
        focusable: FilterOrBool = False,
        focus_on_click: FilterOrBool = False,
        scale: float = 0,
        bbox: DiInt | None = None,
    ) -> None:
        """Create a new kitty graphic instance."""
        super().__init__(
            data=data,
            format_=format_,
            path=path,
            fg_color=fg_color,
            bg_color=bg_color,
            sizing_func=sizing_func,
            focusable=focusable,
            focus_on_click=focus_on_click,
            scale=scale,
            bbox=bbox,
        )
        self.kitty_image_id = 0
        self.loaded = False

    def convert_data(self, rows: int, cols: int) -> str:
        """Convert the graphic's data to base64 data for kitty graphics protocol."""
        return str(
            convert(
                self.data,
                from_=self.format_,
                to="base64-png",
                cols=cols,
                rows=rows,
                fg=self.fg_color,
                bg=self.bg_color,
                path=self.path,
            )
        ).replace("\n", "")

    @staticmethod
    def _kitty_cmd(chunk: str = "", **params: Any) -> str:
        param_str = ",".join(
            [f"{key}={value}" for key, value in params.items() if value is not None]
        )
        cmd = f"\x1b_G{param_str}"
        if chunk:
            cmd += f";{chunk}"
        cmd += "\x1b\\"
        return cmd

    def load(self, rows: int, cols: int) -> None:
        """Send the graphic to the terminal without displaying it."""
        global _kitty_image_count

        data = self.convert_data(rows=rows, cols=cols)
        self.kitty_image_id = _kitty_image_count
        _kitty_image_count += 1

        while data:
            chunk, data = data[:4096], data[4096:]
            cmd = self._kitty_cmd(
                chunk=chunk,
                a="t",  # We are sending an image without displaying it
                t="d",  # Transferring the image directly
                i=self.kitty_image_id,  # Send a unique image number, wait for an image id
                # I=self.kitty_image_number,  # Send a unique image number, wait for an image id
                p=1,  # Placement ID
                q=2,  # No chatback
                f=100,  # Sending a PNG image
                C=1,  # Do not move the cursor
                m=1 if data else 0,  # Data will be chunked
            )
            self.app.output.write_raw(tmuxify(cmd))
        self.app.output.flush()
        self.loaded = True

    def hide(self) -> None:
        """Hide the graphic from show without deleting it."""
        if self.kitty_image_id > 0:
            self.app.output.write_raw(
                tmuxify(
                    self._kitty_cmd(
                        a="d",
                        d="i",
                        i=self.kitty_image_id,
                        q=1,
                    )
                )
            )
            self.app.output.flush()

    def delete(self) -> None:
        """Delete the graphic from the terminal."""
        if self.kitty_image_id > 0:
            self.app.output.write_raw(
                tmuxify(
                    self._kitty_cmd(
                        a="D",
                        d="I",
                        i=self.kitty_image_id,
                        q=2,
                    )
                )
            )
            self.app.output.flush()
            self.loaded = False

    def get_rendered_lines(
        self, width: int, height: int, wrap_lines: bool = False
    ) -> list[StyleAndTextTuples]:
        """Get rendered lines from the cache, or generate them."""
        # TODO - wezterm does not scale kitty graphics, so we might want to resize
        # images at this point rather than just loading them once
        if not self.loaded:
            self.load(cols=width, rows=height)

        def render_lines() -> list[StyleAndTextTuples]:
            """Render the lines to display in the control."""
            full_width = width + self.bbox.left + self.bbox.right
            full_height = height + self.bbox.top + self.bbox.bottom
            cmd = self._kitty_cmd(
                a="p",  # Display a previously transmitted image
                i=self.kitty_image_id,
                p=1,  # Placement ID
                m=0,  # No batches remaining
                q=2,  # No backchat
                c=width,
                r=height,
                C=1,  # 1 = Do move the cursor
                # Horizontal pixel offset of the displayed image region
                x=int(self.px * self.bbox.left / full_width),
                # Vertical pixel offset of the displayed image region
                y=int(self.py * self.bbox.top / full_height),
                # Pixel width of the displayed image region
                w=int(self.px * width / full_width),
                # Pixel height of the displayed image region
                h=int(self.py * height / full_height),
                z=-(2**30) - 1,
            )
            return list(
                split_lines(
                    to_formatted_text(
                        [
                            # Move cursor down and acoss by image height and width
                            ("", "\n".join((height) * [" " * (width)])),
                            # Save position, then move back
                            ("[ZeroWidthEscape]", "\x1b[s"),
                            # Move cursor up if there is more than one line to display
                            *(
                                [("[ZeroWidthEscape]", f"\x1b[{height-1}A")]
                                if height > 1
                                else []
                            ),
                            ("[ZeroWidthEscape]", f"\x1b[{width}D"),
                            # Place the image without moving cursor
                            ("[ZeroWidthEscape]", tmuxify(cmd)),
                            # Restore the last known cursor position (at the bottom)
                            ("[ZeroWidthEscape]", "\x1b[u"),
                        ]
                    )
                )
            )

        key = (width, height, self.bbox, self.app.term_info.cell_size_px)
        return self._format_cache.get(key, render_lines)

    def reset(self) -> None:
        """Hide and delete the kitty graphic from the terminal."""
        self.hide()
        self.delete()
        super().reset()

    def close(self) -> None:
        """Remove the displayed object entirely."""
        super().close()
        if not self.app.leave_graphics():
            self.delete()


class NotVisible(Exception):
    """Exception to signal that a graphic is not currently visible."""

    pass


def get_position_func_overlay(
    target_window: Window,
) -> Callable[[Screen], tuple[WritePosition, DiInt]]:
    """Generate function to positioning floats over existing windows."""

    def get_position(screen: Screen) -> tuple[WritePosition, DiInt]:
        target_wp = screen.visible_windows_to_write_positions.get(target_window)
        if target_wp is None:
            raise NotVisible

        render_info = target_window.render_info
        if render_info is None:
            raise NotVisible

        xpos = target_wp.xpos
        ypos = target_wp.ypos

        content_height = render_info.ui_content.line_count
        content_width = target_wp.width  # TODO - get the actual content width

        # Calculate the cropping box in case the window is scrolled
        bbox = DiInt(
            top=render_info.vertical_scroll,
            right=max(
                0,
                content_height
                - target_wp.width
                - getattr(render_info, "horizontal_scroll", 0),
            ),
            bottom=max(
                0,
                content_height - target_wp.height - render_info.vertical_scroll,
            ),
            left=getattr(render_info, "horizontal_scroll", 0),
        )

        # If the target is within a scrolling container, we might need to adjust
        # the position of the cropped region so the float covers only the visible
        # part of the target window
        if isinstance(target_wp, BoundedWritePosition):
            bbox = bbox._replace(
                top=bbox.top + target_wp.bbox.top,
                right=bbox.right + target_wp.bbox.right,
                bottom=bbox.bottom + target_wp.bbox.bottom,
                left=bbox.left + target_wp.bbox.left,
            )
            xpos += bbox.left
            ypos += bbox.top
            content_height -= bbox.top + bbox.bottom
            content_width -= bbox.left + bbox.right

        new_write_position = WritePosition(
            xpos=xpos,
            ypos=ypos,
            width=max(0, content_width),
            height=max(0, content_height),
        )

        return new_write_position, bbox

    return get_position


class GraphicWindow(Window):
    """A window responsible for displaying terminal graphics content.

    The content is displayed floating on top of a target window.

    The graphic will be displayed if:
    - a completion menu is not being shown
    - a dialog is not being shown
    - a menu is not being shown
    - the output it attached to is fully in view
    """

    content: GraphicControl

    def __init__(
        self,
        content: GraphicControl,
        position: Callable[[Screen], tuple[WritePosition, DiInt]],
        filter: FilterOrBool = True,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initiate a new :py:class:`GraphicWindow` object.

        Args:
            content: A control which generates the graphical content to display
            position: A function which returns the position of the graphic
            filter: A filter which determines if the graphic should be shown
            args: Positional arguments for :py:method:`Window.__init__`
            kwargs: Key-word arguments for :py:method:`Window.__init__`
        """
        super().__init__(*args, **kwargs)
        self.content = content
        self.position = position
        self.filter = ~has_completions & ~has_dialog & ~has_menus & to_filter(filter)

    def write_to_screen(
        self,
        screen: Screen,
        mouse_handlers: MouseHandlers,
        write_position: WritePosition,
        parent_style: str,
        erase_bg: bool,
        z_index: int | None,
    ) -> None:
        """Draw the graphic window's contents to the screen if required."""
        filter_value = self.filter()
        if filter_value:
            try:
                new_write_position, bbox = self.position(screen)
            except NotVisible:
                pass
            else:
                # Inform the graphic control of the calculated cropping region
                self.content.bbox = bbox

                # Draw the graphic content to the screen
                if (
                    new_write_position
                    and new_write_position.width
                    and new_write_position.height
                ):
                    super().write_to_screen(
                        screen,
                        MouseHandlers(),  # Do not let the float add mouse events
                        new_write_position,
                        # Force renderer refreshes by constantly changing the style
                        f"{parent_style} class:render-{get_app().render_counter}",
                        erase_bg=True,
                        z_index=z_index,
                    )
                    return

        # Otherwise hide the content (required for kitty graphics)
        if not filter_value or not get_app().leave_graphics():
            self.content.hide()

    def _fill_bg(
        self,
        screen: Screen,
        # mouse_handlers: MouseHandlers,
        write_position: WritePosition,
        erase_bg: bool,
    ) -> None:
        """Erase/fill the background."""
        char: str | None
        if callable(self.char):
            char = self.char()
        else:
            char = self.char
        if erase_bg or char:
            wp = write_position
            char_obj = Char(char or " ", f"class:render-{get_app().render_counter}")

            # mouse_handlers_dict = mouse_handlers.mouse_handlers
            for y in range(wp.ypos, wp.ypos + wp.height):
                row = screen.data_buffer[y]
                # mouse_handler_row = mouse_handlers_dict[y]
                for x in range(wp.xpos, wp.xpos + wp.width):
                    row[x] = char_obj
                    # mouse_handler_row[x] = lambda e: NotImplemented


def select_graphic_control(format_: str) -> type[GraphicControl] | None:
    """Determine which graphic control to use."""
    SelectedGraphicControl: type[GraphicControl] | None = None
    app = get_app()
    term_info = app.term_info
    preferred_graphics_protocol = app.config.graphics
    useable_graphics_controls: list[type[GraphicControl]] = []
    _in_tmux = in_tmux()

    if preferred_graphics_protocol != "none":
        if term_info.iterm_graphics_status.value and find_route(format_, "base64-png"):
            useable_graphics_controls.append(ItermGraphicControl)
        if (
            preferred_graphics_protocol == "iterm"
            and ItermGraphicControl in useable_graphics_controls
            # Iterm does not work in tmux without pass-through
            and (not _in_tmux or (_in_tmux and app.config.tmux_graphics))
        ):
            SelectedGraphicControl = ItermGraphicControl
        elif (
            term_info.kitty_graphics_status.value
            and find_route(format_, "base64-png")
            # Kitty does not work in tmux without pass-through
            and (not _in_tmux or (_in_tmux and app.config.tmux_graphics))
        ):
            useable_graphics_controls.append(KittyGraphicControl)
        if (
            preferred_graphics_protocol == "kitty"
            and KittyGraphicControl in useable_graphics_controls
        ):
            SelectedGraphicControl = KittyGraphicControl
        # Tmux now supports sixels (>=3.4)
        elif term_info.sixel_graphics_status.value and find_route(format_, "sixel"):
            useable_graphics_controls.append(SixelGraphicControl)
        if (
            preferred_graphics_protocol == "sixel"
            and SixelGraphicControl in useable_graphics_controls
        ):
            SelectedGraphicControl = SixelGraphicControl

        if SelectedGraphicControl is None and useable_graphics_controls:
            SelectedGraphicControl = useable_graphics_controls[0]

    return SelectedGraphicControl


class Display:
    """Rich output displays.

    A container for displaying rich output data.

    """

    def __init__(
        self,
        data: Any,
        format_: str,
        path: Path | None = None,
        fg_color: str | None = None,
        bg_color: str | None = None,
        height: AnyDimension = None,
        width: AnyDimension = None,
        px: int | None = None,
        py: int | None = None,
        focusable: FilterOrBool = False,
        focus_on_click: FilterOrBool = False,
        wrap_lines: FilterOrBool = False,
        always_hide_cursor: FilterOrBool = True,
        scrollbar: FilterOrBool = True,
        scrollbar_autohide: FilterOrBool = True,
        dont_extend_height: FilterOrBool = True,
        style: str | Callable[[], str] = "",
    ) -> None:
        """Instantiate an Output container object.

        Args:
            data: Raw cell output data
            format_: The conversion format of the data to render
            path: The path to the data's original location
            fg_color: The foreground colour to use when renderin this output
            bg_color: The background colour to use when renderin this output
            height: The height of the output
            width: The width of the output
            px: The pixel width of the data if known
            py: The pixel height of the data if known
            focusable: If the output should be focusable
            focus_on_click: If the output should become focused when clicked
            wrap_lines: If the output's lines should be wrapped
            always_hide_cursor: When true, the cursor is never shown
            scrollbar: Whether to show a scrollbar
            scrollbar_autohide: Whether to automatically hide the scrollbar
            dont_extend_height: Whether the window should fill the available height
            style: The style to apply to the output

        """
        self._style = style
        self.fg_color = fg_color
        self.bg_color = bg_color

        # Get data pixel dimensions
        self._px = px
        self._py = py

        sizing_func = self.make_sizing_func(
            data=data,
            format_=format_,
            fg=fg_color,
            bg=bg_color,
        )

        self.control = FormattedTextDisplayControl(
            data,
            format_=format_,
            path=path,
            fg_color=fg_color,
            bg_color=bg_color,
            sizing_func=sizing_func,
            focusable=focusable,
            focus_on_click=focus_on_click,
            wrap_lines=wrap_lines,
        )

        self.window = DisplayWindow(
            content=self.control,
            height=height,
            width=width,
            wrap_lines=False,
            always_hide_cursor=always_hide_cursor,
            dont_extend_height=dont_extend_height,
            style=self.style,
            char=" ",
        )

        self.container = VSplit(
            [
                self.window,
                ConditionalContainer(
                    MarginContainer(ScrollbarMargin(), target=self.window),
                    filter=to_filter(scrollbar)
                    & (
                        ~to_filter(scrollbar_autohide)
                        | (to_filter(scrollbar_autohide) & scrollable(self.window))
                    ),
                ),
            ]
        )

        # Add graphic
        self.graphic_control = None
        if SelectedGraphicControl := select_graphic_control(format_):
            self.graphic_control = SelectedGraphicControl(
                data,
                format_=format_,
                path=path,
                fg_color=fg_color,
                bg_color=bg_color,
                sizing_func=sizing_func,
            )
            graphic_window = GraphicWindow(
                content=self.graphic_control,
                position=get_position_func_overlay(self.window),
                style=self.style,
            )

            # The only reference to the float is saved on the Display widget
            self.graphic_float = Float(content=graphic_window)
            # Hide the graphic if not in the app's list of graphics
            weak_float_ref = weakref.ref(self.graphic_float)
            graphic_window.filter &= Condition(
                lambda: weak_float_ref() in get_app().graphics
            )
            # Hide the graphic if the float is deleted
            weakref.finalize(self.graphic_float, self.graphic_control.close)
            # Add graphic to app
            get_app().graphics.add(self.graphic_float)

    def style(self) -> str:
        """Use the background color of the data as the default style."""
        style = to_str(self._style)
        if self.bg_color:
            style = f"bg:{self.bg_color} {style}"
        return style

    @property
    def data(self) -> Any:
        """Return the display's current data."""
        return self.control.data

    @data.setter
    def data(self, value: Any) -> None:
        """Set the display container's data."""
        self.control.data = value
        if self.graphic_control is not None:
            self.graphic_control.data = value

    @property
    def format_(self) -> str:
        """Return the display's current data format."""
        return self.control.format_

    @format_.setter
    def format_(self, value: str) -> None:
        """Set the display container's data format."""
        self.control.format_ = value
        if self.graphic_control is not None:
            self.graphic_control.format_ = value

    @property
    def path(self) -> Path | None:
        """Return the display's current data path."""
        return self.control.path

    @path.setter
    def path(self, value: Path | None) -> None:
        """Set the display container's data path."""
        self.control.path = value
        if self.graphic_control is not None:
            self.graphic_control.path = value

    def make_sizing_func(
        self, data: Any, format_: str, fg: str | None, bg: str | None
    ) -> Callable[[], tuple[int, float]]:
        """Create a function to recalculate the data's dimensions in terminal cells."""
        px, py = self.px, self.py
        if px is None or py is None:
            px, py = data_pixel_size(data, format_, fg=fg, bg=bg)
        return partial(pixels_to_cell_size, px, py)

    @property
    def px(self) -> int | None:
        """Return the displayed data's pixel widget."""
        return self._px

    @px.setter
    def px(self, value: int | None) -> None:
        """Set the display container's width in pixels."""
        self._px = value
        self.update_sizing()

    @property
    def py(self) -> int | None:
        """Return the displayed data's pixel height."""
        return self._py

    @py.setter
    def py(self, value: int | None) -> None:
        """Set the display container's height in pixels."""
        self._py = value
        self.update_sizing()

    def update_sizing(self) -> None:
        """Create a sizing function when the data's pixel size changes."""
        sizing_func = self.make_sizing_func(
            data=self.control.data,
            format_=self.control.format_,
            fg=self.control.fg_color,
            bg=self.control.bg_color,
        )
        self.control.sizing_func = sizing_func
        self.control.reset()
        if self.graphic_control is not None:
            self.graphic_control.sizing_func = sizing_func
            self.graphic_control.reset()

    def __pt_container__(self) -> AnyContainer:
        """Return the content of this output."""
        return self.container

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _scroll_display_left() -> None:
        """Scroll the display up one line."""
        window = get_app().layout.current_window
        assert isinstance(window, DisplayWindow)
        window._scroll_left()

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _scroll_display_right() -> None:
        """Scroll the display down one line."""
        window = get_app().layout.current_window
        assert isinstance(window, DisplayWindow)
        window._scroll_right()

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _scroll_display_up() -> None:
        """Scroll the display up one line."""
        get_app().layout.current_window._scroll_up()

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _scroll_display_down() -> None:
        """Scroll the display down one line."""
        get_app().layout.current_window._scroll_down()

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _page_up_display() -> None:
        """Scroll the display up one page."""
        window = get_app().layout.current_window
        if window.render_info is not None:
            for _ in range(window.render_info.window_height):
                window._scroll_up()

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _page_down_display() -> None:
        """Scroll the display down one page."""
        window = get_app().layout.current_window
        if window.render_info is not None:
            for _ in range(window.render_info.window_height):
                window._scroll_down()

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _go_to_start_of_display() -> None:
        """Scroll the display to the top."""
        from euporie.core.widgets.display import DisplayControl

        current_control = get_app().layout.current_control
        if isinstance(current_control, DisplayControl):
            current_control.cursor_position = Point(0, 0)

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _go_to_end_of_display() -> None:
        """Scroll the display down one page."""
        from euporie.core.widgets.display import DisplayControl

        layout = get_app().layout
        current_control = layout.current_control
        window = layout.current_window
        if (
            isinstance(current_control, DisplayControl)
            and window.render_info is not None
        ):
            current_control.cursor_position = Point(
                0, window.render_info.ui_content.line_count - 1
            )

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.core.widgets.display.Display": {
                "scroll-display-left": "left",
                "scroll-display-right": "right",
                "scroll-display-up": ["up", "k"],
                "scroll-display-down": ["down", "j"],
                "page-up-display": "pageup",
                "page-down-display": "pagedown",
                "go-to-start-of-display": "home",
                "go-to-end-of-display": "end",
            }
        }
    )



"""Contain the main class for a notebook file."""

from __future__ import annotations

import logging
from functools import partial
from typing import TYPE_CHECKING, cast

import nbformat
from prompt_toolkit.application.run_in_terminal import in_terminal
from prompt_toolkit.buffer import Buffer, ValidationState
from prompt_toolkit.filters.app import (
    buffer_has_focus,
    has_completions,
    has_focus,
    has_selection,
    in_paste_mode,
    renderer_height_is_known,
)
from prompt_toolkit.filters.base import Condition
from prompt_toolkit.key_binding.key_bindings import KeyBindings
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    FloatContainer,
    HSplit,
    VSplit,
    Window,
)
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.layout import Layout
from upath import UPath

from euporie.core.commands import add_cmd, get_cmd
from euporie.core.config import add_setting
from euporie.core.filters import (
    at_end_of_buffer,
    buffer_is_code,
    buffer_is_empty,
    kernel_tab_has_focus,
)
from euporie.core.format import format_code
from euporie.core.kernel import MsgCallbacks
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)
from euporie.core.style import KERNEL_STATUS_REPR
from euporie.core.tabs.base import KernelTab
from euporie.core.terminal import edit_in_editor
from euporie.core.validation import KernelValidator
from euporie.core.widgets.cell_outputs import CellOutputArea
from euporie.core.widgets.inputs import KernelInput, StdInput
from euporie.core.widgets.page import PrintingContainer
from euporie.core.widgets.pager import PagerState

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any, Callable, Sequence

    from prompt_toolkit.application.application import Application
    from prompt_toolkit.formatted_text import AnyFormattedText, StyleAndTextTuples
    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent
    from prompt_toolkit.layout.containers import Float

    from euporie.core.app import BaseApp

log = logging.getLogger(__name__)


class Console(KernelTab):
    """Interactive console.

    An interactive console which connects to a Jupyter kernel.

    """

    bg_init = False

    def __init__(
        self,
        app: BaseApp,
        path: Path | None = None,
        use_kernel_history: bool = True,
        connection_file: str = "",
    ) -> None:
        """Create a new :py:class:`Console` tab instance.

        Args:
            app: The euporie application the console tab belongs to
            path: A file path to open (not used currently)
            use_kernel_history: If :const:`True`, history will be loaded from the kernel
            connection_file: The connection file of an existing kernel
        """
        # Kernel setup
        self._metadata = {}
        self.kernel_name = app.config.kernel_name
        self.allow_stdin = True
        self.default_callbacks = MsgCallbacks(
            get_input=lambda prompt, password: self.stdin_box.get_input(
                prompt, password
            ),
            set_execution_count=partial(setattr, self, "execution_count"),
            add_output=self.new_output,
            clear_output=self.clear_output,
            # set_metadata=self.misc_callback,
            set_status=lambda status: self.app.invalidate(),
            set_kernel_info=self.set_kernel_info,
            done=self.complete,
            dead=self.kernel_died,
            set_next_input=self.set_next_input,
            edit_magic=edit_in_editor,
        )
        self.kernel_tab = self

        super().__init__(
            app=app,
            path=path,
            use_kernel_history=use_kernel_history,
            connection_file=app.config.connection_file,
        )

        self.lang_info: dict[str, Any] = {}
        self.execution_count = 0
        self.clear_outputs_on_output = False

        self.json = nbformat.v4.new_notebook()
        self.json["metadata"] = self._metadata

        self.container = self.load_container()

        self.kernel.start(cb=self.kernel_started, wait=False)

        self.app.before_render += self.render_outputs

    def kernel_died(self) -> None:
        """Call when the kernel dies."""
        log.error("The kernel has died")

    async def load_history(self) -> None:
        """Load kernel history."""
        await super().load_history()
        # Re-run history load for the input-box
        self.input_box.buffer._load_history_task = None
        self.input_box.buffer.load_history_if_not_yet_loaded()

    def close(self, cb: Callable | None = None) -> None:
        """Close the console tab."""
        # Ensure any output no longer appears interactive
        self.live_output.style = "class:disabled"
        # Unregister output renderer
        self.app.before_render -= self.render_outputs
        super().close(cb)

    def clear_output(self, wait: bool = False) -> None:
        """Remove the last output, optionally when new output is generated."""
        if wait:
            self.clear_outputs_on_output = True
        else:
            self.live_output.reset()

    def validate_input(self, code: str) -> bool:
        """Determine if the entered code is ready to run."""
        assert self.kernel is not None
        completeness_status = self.kernel.is_complete(code, wait=True).get(
            "status", "unknown"
        )
        if (
            not code.strip()
            or completeness_status == "incomplete"
            or (completeness_status == "unknown" and code[-2:] != "\n\n")
        ):
            return False
        else:
            return True

    def run(self, buffer: Buffer | None = None) -> None:
        """Run the code in the input box."""
        if buffer is None:
            buffer = self.input_box.buffer
        # Auto-reformat code
        if self.app.config.autoformat:
            self.reformat()
        # Get the code to run
        text = buffer.text
        # Remove any selections from input
        self.input_box.buffer.selection_state = None
        # Disable existing output
        self.live_output.style = "class:disabled"
        # Re-render the app and move to below the current output
        self.app.draw()
        # Prevent displayed graphics on terminal being cleaned up (bit of a hack)
        self.app.graphics.clear()
        # Run the previous entry
        if self.kernel.status == "starting":
            self.kernel_queue.append(partial(self.kernel.run, text, wait=False))
        else:
            self.kernel.run(text, wait=False)
        # Increment this for display purposes until we get the response from the kernel
        self.execution_count += 1
        # Reset the input & output
        buffer.reset(append_to_history=True)
        # Remove any live outputs and disable mouse support
        self.live_output.reset()
        if self.app.config.mouse_support is None:
            self.app.need_mouse_support = False
        # Record the input as a cell in the json
        self.json["cells"].append(
            nbformat.v4.new_code_cell(source=text, execution_count=self.execution_count)
        )
        if (
            self.app.config.max_stored_outputs
            and len(self.json["cells"]) > self.app.config.max_stored_outputs
        ):
            del self.json["cells"][0]

    def new_output(self, output_json: dict[str, Any]) -> None:
        """Print the previous output and replace it with the new one."""
        # Clear the output if we were previously asked to
        if self.clear_outputs_on_output:
            self.clear_outputs_on_output = False
            # Clear the screen
            get_cmd("clear-screen").run()

        # Add to record
        if self.json["cells"]:
            self.json["cells"][-1]["outputs"].append(output_json)

        # Set output
        if "application/vnd.jupyter.widget-view+json" in output_json.get("data", {}):
            # Use a live output to display widgets
            self.live_output.add_output(output_json)
            # Enable mouse support if we have a live output
            if self.app.config.mouse_support is None:
                self.app.need_mouse_support = True
        else:
            # Queue the output json
            self.output.add_output(output_json)
            # Invalidate the app so the output get printed
            self.app.invalidate()

    def render_outputs(self, app: Application[Any]) -> None:
        """Request that any unrendered outputs be rendered."""
        if self.output.json:
            self.app.create_background_task(self.async_render_outputs())

    async def async_render_outputs(self) -> None:
        """Render any unrendered outputs above the application."""
        if self.output.json:
            # Run the output app in the terminal
            async with in_terminal():
                self.app.renderer.render(self.app, self.output_layout, is_done=True)
            # Remove the outputs so they do not get rendered again
            self.output.reset()

    def reset(self) -> None:
        """Reet the state of the tab."""
        self.live_output.reset()
        if self.app.config.mouse_support is None:
            self.app.need_mouse_support = False

    def complete(self, content: dict | None = None) -> None:
        """Re-render any changes."""
        self.app.invalidate()

    def prompt(
        self, text: str, offset: int = 0, show_busy: bool = False
    ) -> StyleAndTextTuples:
        """Determine what should be displayed in the prompt of the cell."""
        prompt = str(self.execution_count + offset)
        if show_busy and self.kernel.status in ("busy", "queued"):
            prompt = "*".center(len(prompt))
        ft: StyleAndTextTuples = [
            ("", f"{text}["),
            ("class:count", prompt),
            ("", "]: "),
        ]
        return ft

    @property
    def language(self) -> str:
        """The language of the current kernel."""
        return self.lang_info.get(
            "name", self.lang_info.get("pygments_lexer", "python")
        )

    def lang_file_ext(self) -> str:
        """Return the file extension for scripts in the notebook's language."""
        return self.lang_info.get("file_extension", ".py")

    def load_container(self) -> HSplit:
        """Build the main application layout."""
        # Output area

        self.output = CellOutputArea([], parent=self)

        @Condition
        def first_output() -> bool:
            """Check if the current outputs contain the first output."""
            if self.output.json:
                for output in self.json["cells"][-1].get("outputs", []):
                    if output in self.live_output.json:
                        continue
                    return output == self.output.json[0]
            return False

        output_prompt = Window(
            FormattedTextControl(partial(self.prompt, "Out", show_busy=False)),
            dont_extend_width=True,
            style="class:cell,output,prompt",
            height=1,
        )
        output_margin = Window(
            char=" ", width=lambda: len(str(self.execution_count)) + 7
        )

        output_area = PrintingContainer(
            [
                ConditionalContainer(
                    Window(height=1, dont_extend_height=True), filter=first_output
                ),
                VSplit(
                    [
                        ConditionalContainer(output_prompt, filter=first_output),
                        ConditionalContainer(output_margin, filter=~first_output),
                        self.output,
                    ]
                ),
            ],
        )
        # We need to ensure the output layout also has the application's floats so that
        # graphics are displayed
        self.output_layout = Layout(
            FloatContainer(
                output_area,
                floats=cast("list[Float]", self.app.graphics),
            )
        )

        # Live output area

        self.live_output = CellOutputArea([], parent=self)

        # Input area

        def on_cursor_position_changed(buf: Buffer) -> None:
            """Respond to cursor movements."""
            # Update contextual help
            if self.app.config.autoinspect and buf.name == "code":
                self.inspect()
            elif (pager := self.app.pager) is not None and pager.visible():
                pager.hide()

        input_kb = KeyBindings()

        @Condition
        def empty() -> bool:
            from euporie.console.app import get_app

            buffer = get_app().current_buffer
            text = buffer.text
            return not text.strip()

        @Condition
        def not_invalid() -> bool:
            from euporie.console.app import get_app

            buffer = get_app().current_buffer
            return buffer.validation_state != ValidationState.INVALID

        @input_kb.add(
            "enter",
            filter=has_focus("code")
            & ~empty
            & not_invalid
            & at_end_of_buffer
            & ~has_completions,
        )
        async def on_enter(event: KeyPressEvent) -> NotImplementedOrNone:
            """Accept input if the input is valid, otherwise insert a return."""
            buffer = event.current_buffer
            # Accept the buffer if there are 2 blank lines
            accept = buffer.text[-2:] == "\n\n"
            # Also accept if the input is valid
            if not accept:
                accept = buffer.validate(set_cursor=False)
            if accept:
                if buffer.accept_handler:
                    keep_text = buffer.accept_handler(buffer)
                else:
                    keep_text = False
                # buffer.append_to_history()
                if not keep_text:
                    buffer.reset()
                return NotImplemented

            # Process the input as a regular :kbd:`enter` key-press
            event.key_processor.feed(event.key_sequence[0], first=True)
            # Prevent the app getting invalidated
            return None

        @input_kb.add("s-enter")
        def _newline(event: KeyPressEvent) -> None:
            """Force new line on Shift-Enter."""
            event.current_buffer.newline(copy_margin=not in_paste_mode())

        input_prompt = Window(
            FormattedTextControl(partial(self.prompt, "In ", offset=1)),
            dont_extend_width=True,
            style="class:cell,input,prompt",
            height=1,
        )

        self.input_box = KernelInput(
            kernel_tab=self,
            accept_handler=self.run,
            on_cursor_position_changed=on_cursor_position_changed,
            validator=KernelValidator(self.kernel),
            # validate_while_typing=False,
            enable_history_search=True,
            key_bindings=input_kb,
        )
        self.input_box.buffer.name = "code"

        self.app.focused_element = self.input_box.buffer

        self.stdin_box = StdInput(self)

        self.input_layout = HSplit(
            [
                ConditionalContainer(
                    HSplit(
                        [
                            Window(height=1, dont_extend_height=True),
                            VSplit([output_margin, self.live_output]),
                        ]
                    ),
                    filter=Condition(lambda: bool(self.live_output.json)),
                ),
                # StdIn
                self.stdin_box,
                ConditionalContainer(
                    Window(height=1, dont_extend_height=True),
                    filter=self.stdin_box.visible,
                ),
                # Spacing
                ConditionalContainer(
                    Window(height=1, dont_extend_height=True),
                    filter=Condition(lambda: self.execution_count > 0)
                    & (
                        (
                            renderer_height_is_known
                            & Condition(lambda: self.app.renderer.rows_above_layout > 0)
                        )
                        | ~renderer_height_is_known
                    ),
                ),
                # Input
                ConditionalContainer(
                    VSplit(
                        [
                            input_prompt,
                            self.input_box,
                        ],
                    ),
                    filter=~self.stdin_box.visible,
                ),
            ],
            key_bindings=load_registered_bindings(
                "euporie.console.tabs.console.Console"
            ),
        )

        return self.input_layout

    def set_next_input(self, text: str, replace: bool = False) -> None:
        """Set the text for the next prompt."""
        self.input_box.buffer.text = text

    def accept_stdin(self, buf: Buffer) -> bool:
        """Accept the user's input."""
        return True

    def interrupt_kernel(self) -> None:
        """Interrupt the current `Notebook`'s kernel."""
        assert self.kernel is not None
        self.kernel.interrupt()

    def set_kernel_info(self, info: dict) -> None:
        """Receive and processes kernel metadata."""
        self.lang_info = info.get("language_info", {})

    def refresh(self, now: bool = True) -> None:
        """Request the output is refreshed (does nothing)."""
        pass

    def __pt_status__(
        self,
    ) -> tuple[Sequence[AnyFormattedText], Sequence[AnyFormattedText]]:
        """Generate the formatted text for the statusbar."""
        assert self.kernel is not None
        return (
            [],
            [
                [
                    (
                        "",
                        self.kernel_display_name
                        # , self._statusbar_kernel_handeler)],
                    )
                ],
                KERNEL_STATUS_REPR.get(self.kernel.status, "."),
            ],
        )

    def reformat(self) -> None:
        """Reformat the input."""
        self.input_box.text = format_code(self.input_box.text, self.app.config)

    def inspect(self) -> None:
        """Get contextual help for the current cursor position in the current cell."""
        code = self.input_box.text
        cursor_pos = self.input_box.buffer.cursor_position

        assert self.app.pager is not None

        if self.app.pager.visible() and self.app.pager.state is not None:
            if (
                self.app.pager.state.code == code
                and self.app.pager.state.cursor_pos == cursor_pos
            ):
                self.app.pager.focus()
                return

        def _cb(response: dict) -> None:
            assert self.app.pager is not None
            prev_state = self.app.pager.state
            new_state = PagerState(
                code=code,
                cursor_pos=cursor_pos,
                response=response,
            )
            if prev_state != new_state:
                self.app.pager.state = new_state
                self.app.invalidate()

        assert self.kernel is not None
        self.kernel.inspect(
            code=code,
            cursor_pos=cursor_pos,
            callback=_cb,
        )

    def save(self, path: Path | None = None, cb: Callable | None = None) -> None:
        """Save the console as a notebook."""
        from euporie.core.tabs.notebook import BaseNotebook

        if path is not None:
            BaseNotebook.save(cast("BaseNotebook", self), path)

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd()
    def _accept_input() -> None:
        """Accept the current console input."""
        from euporie.console.app import get_app

        buffer = get_app().current_buffer
        if buffer:
            buffer.validate_and_handle()

    @staticmethod
    @add_cmd(
        filter=buffer_is_code & buffer_has_focus & ~has_selection & ~buffer_is_empty,
    )
    def _clear_input() -> None:
        """Clear the console input."""
        from euporie.console.app import get_app

        buffer = get_app().current_buffer
        buffer.reset()

    @staticmethod
    @add_cmd(
        filter=buffer_is_code & buffer_has_focus,
    )
    def _run_input() -> None:
        """Run the console input."""
        from euporie.console.app import get_app

        console = get_app().tab
        assert isinstance(console, Console)
        console.run()

    @staticmethod
    @add_cmd(
        filter=buffer_is_code & buffer_has_focus & ~has_selection,
    )
    def _show_contextual_help() -> None:
        """Display contextual help."""
        from euporie.console.app import get_app

        console = get_app().tab
        assert isinstance(console, Console)
        console.inspect()

    @staticmethod
    @add_cmd(
        name="cc-interrupt-kernel",
        hidden=True,
        filter=buffer_is_code & buffer_is_empty,
    )
    @add_cmd(filter=kernel_tab_has_focus)
    def _interrupt_kernel() -> None:
        """Interrupt the notebook's kernel."""
        from euporie.console.app import get_app

        if isinstance(kt := get_app().tab, KernelTab):
            kt.interrupt_kernel()

    @staticmethod
    @add_cmd(filter=kernel_tab_has_focus)
    def _restart_kernel() -> None:
        """Restart the notebook's kernel."""
        from euporie.console.app import get_app

        if isinstance(kt := get_app().tab, KernelTab):
            kt.restart_kernel()

    # ################################### Settings ####################################

    add_setting(
        name="max_stored_outputs",
        flags=["--max-stored-outputs"],
        type_=int,
        help_="The number of inputs / outputs to store in an in-memory notebook",
        default=100,
        schema={
            "minimum": 0,
        },
        description="""
            Defines the maximum number of executed "cells" to store in case the console
            session is saved to a file or converted into a notebook.
        """,
    )

    add_setting(
        name="connection_file",
        flags=["--connection-file", "--kernel-connection-file"],
        type_=UPath,
        help_="Attempt to connect to an existing kernel using a JSON connection info file",
        default=None,
        description="""
            If the file does not exist, kernel connection information will be written
            to the file path provided.

            If the file exists, kernel connection info will be read from the file,
            allowing euporie to connect to existing kernels.
        """,
    )

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.console.tabs.console.Console": {
                "cc-interrupt-kernel": "c-c",
            },
            "euporie.console.app.ConsoleApp": {
                "clear-input": "c-c",
                "run-input": ["c-enter", "c-e"],
            },
        }
    )



"""Suggest line completions from kernel history."""

from __future__ import annotations

import logging
from collections import deque
from typing import TYPE_CHECKING

from prompt_toolkit.auto_suggest import AutoSuggest, ConditionalAutoSuggest, Suggestion
from prompt_toolkit.filters import to_filter

if TYPE_CHECKING:
    from prompt_toolkit.buffer import Buffer
    from prompt_toolkit.document import Document
    from prompt_toolkit.filters import Filter
    from prompt_toolkit.history import History

    from euporie.core.kernel import Kernel

log = logging.getLogger(__name__)


class HistoryAutoSuggest(AutoSuggest):
    """Suggest line completions from a :class:`History` object."""

    def __init__(self, history: History, cache_size: int = 100_000) -> None:
        """Set the kernel instance in initialization."""
        self.history = history

        self.cache_size = cache_size
        self.cache_keys: deque[str] = deque()
        self.cache: dict[str, Suggestion] = {}

    def get_suggestion(self, buffer: Buffer, document: Document) -> Suggestion | None:
        """Get a line completion suggestion."""
        result: Suggestion | None = None
        line = document.current_line.lstrip()
        if line:
            if line in self.cache:
                result = self.cache[line]
            else:
                result = self.lookup_suggestion(line)
                if result:
                    if len(self.cache) > self.cache_size:
                        key_to_remove = self.cache_keys.popleft()
                        if key_to_remove in self.cache:
                            del self.cache[key_to_remove]

                    self.cache_keys.append(line)
                    self.cache[line] = result
        return result

    def lookup_suggestion(self, line: str) -> Suggestion | None:
        """Find the most recent matching line in the history."""
        # Loop history, most recent item first
        for text in self.history._loaded_strings:
            if line in text:
                # Loop over lines of item in reverse order
                for hist_line in text.splitlines()[::-1]:
                    hist_line = hist_line.strip()
                    if hist_line.startswith(line):
                        # Return from the match to end from the history line
                        suggestion = hist_line[len(line) :]
                        return Suggestion(suggestion)
        return None

    '''
    def lookup_suggestion(self, line: "str") -> "Optional[Suggestion]":
         ""Suggest most commonly used line."""
        results = {}
        pat = re.compile(f"^\\s*{re.escape(line)}(.*)$", re.MULTILINE)
        for text in self.history.get_strings():
            if line in text:
                for hist_line in text.splitlines()[::-1]:
                    hist_line = hist_line.strip()
                    if hist_line.startswith(line):
                        # Return from the match to end from the history line
                        suggestion = hist_line[len(line) :]
                        if suggestion not in results:
                            results[suggestion] = 1
                        else:
                            results[suggestion] += 1
        return Suggestion(
            {b: a for a, b in results.items()}[max(results.values())]
        )
    '''


class KernelAutoSuggest(AutoSuggest):
    """Suggest line completions from kernel history."""

    def __init__(self, kernel: Kernel) -> None:
        """Set the kernel instance in initialization."""
        self.kernel = kernel

    def get_suggestion(self, buffer: Buffer, document: Document) -> Suggestion | None:
        """Doe nothing."""
        return None

    async def get_suggestion_async(
        self, buff: Buffer, document: Document
    ) -> Suggestion | None:
        """Return suggestions based on matching kernel history."""
        line = document.current_line.strip()
        if line:
            suggestions = await self.kernel.history_(f"*{line}*")
            log.debug("Suggestor got suggestions %s", suggestions)
            if suggestions:
                _, _, text = suggestions[0]
                # Find matching line
                for hist_line in text.split("\n"):
                    hist_line = hist_line.strip()
                    if hist_line.startswith(line):
                        # Return from the match to end from the history line
                        suggestion = hist_line[len(line) :]
                        log.debug("Suggesting %s", suggestion)
                        return Suggestion(suggestion)
        return None


class ConditionalAutoSuggestAsync(ConditionalAutoSuggest):
    """Auto suggest that can be turned on and of according to a certain condition."""

    def __init__(self, auto_suggest: AutoSuggest, filter: bool | Filter) -> None:
        """Create a new asynchronous conditional autosuggestion wrapper.

        Args:
            auto_suggest: The :class:`AutoSuggest` to use to retrieve suggestions
            filter: The filter use to determine if autosuggestions should be retrieved

        """
        self.auto_suggest = auto_suggest
        self.filter = to_filter(filter)

    async def get_suggestion_async(
        self, buffer: Buffer, document: Document
    ) -> Suggestion | None:
        """Get suggestions asynchronously if the filter allows."""
        if self.filter():
            return await self.auto_suggest.get_suggestion_async(buffer, document)

        return None



"""Custom validators."""

from __future__ import annotations

from typing import TYPE_CHECKING

from prompt_toolkit.validation import ValidationError, Validator

if TYPE_CHECKING:
    from prompt_toolkit.document import Document

    from euporie.core.kernel import Kernel


class KernelValidator(Validator):
    """Validate kernel input using a kernel code completeness call."""

    def __init__(self, kernel: Kernel) -> None:
        """Initialize the validator."""
        self.kernel = kernel

    def validate(self, document: Document) -> None:
        """Validate the input synchronously."""
        completeness_status = self.kernel.is_complete(
            code=document.text, wait=True
        ).get("status", "unknown")
        if completeness_status == "incomplete":
            raise ValidationError

    async def validate_async(self, document: Document) -> None:
        """Return a `Future` which is set when the validation is ready."""
        completeness_status = (await self.kernel.is_complete_(code=document.text)).get(
            "status", "unknown"
        )
        if completeness_status == "incomplete":
            raise ValidationError
        return None



"""Decorative widgets."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from prompt_toolkit.cache import FastDictCache
from prompt_toolkit.filters import has_focus, to_filter
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    Container,
    DynamicContainer,
    Float,
    FloatContainer,
    HSplit,
    VSplit,
    Window,
    to_container,
)
from prompt_toolkit.layout.dimension import Dimension
from prompt_toolkit.layout.mouse_handlers import MouseHandlers
from prompt_toolkit.layout.screen import Char, Screen, WritePosition
from prompt_toolkit.mouse_events import MouseEventType

from euporie.core.border import ThinLine
from euporie.core.config import add_setting
from euporie.core.current import get_app
from euporie.core.data_structures import DiBool
from euporie.core.style import ColorPaletteColor

if TYPE_CHECKING:
    from typing import Callable

    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.layout.containers import AnyContainer
    from prompt_toolkit.mouse_events import MouseEvent

    from euporie.core.border import GridStyle

    MouseHandler = Callable[[MouseEvent], object]

log = logging.getLogger(__name__)


class Line(Container):
    """Draw a horizontal or vertical line."""

    def __init__(
        self,
        char: str | None = None,
        width: int | None = None,
        height: int | None = None,
        collapse: bool = False,
        style: str = "class:grid-line",
    ) -> None:
        """Initialize a grid line.

        Args:
            char: The character to draw. If unset, the relevant character from
                :py:class:`euporie.core.box.grid` is used
            width: The length of the line. If specified, the line will be horizontal
            height: The height of the line. If specified, the line will be vertical
            collapse: Whether to hide the line when there is not enough space
            style: Style to apply to the line

        Raises:
            ValueError: If both width and height are specified. A line must only have a
                single dimension.

        """
        if width and height:
            raise ValueError("Only one of `width` or `height` must be set")
        self.width = width
        self.height = height
        if char is None:
            char = ThinLine.grid.VERTICAL if width else ThinLine.grid.HORIZONTAL
        self.char = Char(char, style)
        self.collapse = collapse

    def reset(self) -> None:
        """Reet the state of the line. Does nothing."""

    def preferred_width(self, max_available_width: int) -> Dimension:
        """Return the preferred width of the line."""
        return Dimension(min=int(not self.collapse), max=self.width)

    def preferred_height(self, width: int, max_available_height: int) -> Dimension:
        """Return the preferred height of the line."""
        return Dimension(min=int(not self.collapse), max=self.height)

    def write_to_screen(
        self,
        screen: Screen,
        mouse_handlers: MouseHandlers,
        write_position: WritePosition,
        parent_style: str,
        erase_bg: bool,
        z_index: int | None,
    ) -> None:
        """Draw a continuous line in the ``write_position`` area.

        Args:
            screen: The :class:`~prompt_toolkit.layout.screen.Screen` class to which
                the output has to be written.
            mouse_handlers: :class:`prompt_toolkit.layout.mouse_handlers.MouseHandlers`.
            write_position: A :class:`prompt_toolkit.layout.screen.WritePosition` object
                defining where this container should be drawn.
            erase_bg: If true, the background will be erased prior to drawing.
            parent_style: Style string to pass to the :class:`.Window` object. This will
                be applied to all content of the windows. :class:`.VSplit` and
                :class:`prompt_toolkit.layout.containers.HSplit` can use it to pass
                their style down to the windows that they contain.
            z_index: Used for propagating z_index from parent to child.

        """
        ypos = write_position.ypos
        xpos = write_position.xpos

        for y in range(ypos, ypos + write_position.height):
            row = screen.data_buffer[y]
            for x in range(xpos, xpos + write_position.width):
                row[x] = self.char

    def get_children(self) -> list:
        """Return an empty list of the container's children."""
        return []


class Pattern(Container):
    """Fill an area with a repeating background pattern."""

    def __init__(
        self,
        char: str | Callable[[], str],
        pattern: int | Callable[[], int] = 1,
    ) -> None:
        """Initialize the :class:`Pattern`."""
        self.bg = Char(" ", "class:pattern")
        self.char = char
        self.pattern = pattern

    def reset(self) -> None:
        """Reet the pattern. Does nothing."""
        pass

    def preferred_width(self, max_available_width: int) -> Dimension:
        """Return an empty dimension (expand to available width)."""
        return Dimension()

    def preferred_height(self, width: int, max_available_height: int) -> Dimension:
        """Return an empty dimension (expand to available height)."""
        return Dimension()

    def write_to_screen(
        self,
        screen: Screen,
        mouse_handlers: MouseHandlers,
        write_position: WritePosition,
        parent_style: str,
        erase_bg: bool,
        z_index: int | None,
    ) -> None:
        """Fill the whole area of write_position with a pattern.

        Args:
            screen: The :class:`~prompt_toolkit.layout.screen.Screen` class to which
                the output has to be written.
            mouse_handlers: :class:`prompt_toolkit.layout.mouse_handlers.MouseHandlers`.
            write_position: A :class:`prompt_toolkit.layout.screen.WritePosition` object
                defining where this container should be drawn.
            erase_bg: If true, the background will be erased prior to drawing.
            parent_style: Style string to pass to the :class:`.Window` object. This will
                be applied to all content of the windows. :class:`.VSplit` and
                :class:`prompt_toolkit.layout.containers.HSplit` can use it to pass
                their style down to the windows that they contain.
            z_index: Used for propagating z_index from parent to child.

        """
        bg = self.bg
        if callable(self.pattern):
            pattern = self.pattern()
        else:
            pattern = self.pattern
        if callable(self.char):
            char = Char(self.char(), "class:pattern")
        else:
            char = Char(self.char, "class:pattern")

        ypos = write_position.ypos
        xpos = write_position.xpos

        for y in range(ypos, ypos + write_position.height):
            row = screen.data_buffer[y]
            for x in range(xpos, xpos + write_position.width):
                if (
                    (pattern == 1)
                    or (pattern == 2 and (x + y) % 2 == 0)
                    or (pattern == 3 and (x + 2 * y) % 4 == 0)
                    or (pattern == 4 and (x + y) % 3 == 0)
                    or (pattern == 5 and ((x + y % 2 * 3) % 6) % 4 == 0)
                ):
                    row[x] = char
                else:
                    row[x] = bg

    def get_children(self) -> list:
        """Return an empty list of the container's children."""
        return []


class Border:
    """Draw a border around any container."""

    def __init__(
        self,
        body: AnyContainer,
        border: GridStyle | None = ThinLine.grid,
        style: str | Callable[[], str] = "class:border",
        show_borders: DiBool | None = None,
    ) -> None:
        """Create a new border widget which wraps another container.

        Args:
            body: The container to surround with a border
            border: The grid style to use
            style: The style to apply to the border
            show_borders: Which of the four borders should be displayed

        """
        self.body = body
        self.style = style

        if show_borders:
            show_borders = DiBool(*show_borders)
        else:
            show_borders = DiBool(True, True, True, True)
        border_top = to_filter(show_borders.top)
        border_right = to_filter(show_borders.right)
        border_bottom = to_filter(show_borders.bottom)
        border_left = to_filter(show_borders.left)

        self.container: AnyContainer
        if border is not None and any(show_borders):
            self.container = HSplit(
                [
                    ConditionalContainer(
                        VSplit(
                            [
                                ConditionalContainer(
                                    Window(
                                        width=1,
                                        height=1,
                                        char=border.TOP_LEFT,
                                        style=self.add_style("class:left,top"),
                                    ),
                                    filter=border_top & border_left,
                                ),
                                ConditionalContainer(
                                    Window(
                                        char=border.TOP_MID,
                                        style=self.add_style("class:top"),
                                    ),
                                    filter=border_top,
                                ),
                                ConditionalContainer(
                                    Window(
                                        width=1,
                                        height=1,
                                        char=border.TOP_RIGHT,
                                        style=self.add_style("class:right,top"),
                                    ),
                                    filter=border_top & border_right,
                                ),
                            ],
                            height=1,
                        ),
                        filter=border_top,
                    ),
                    VSplit(
                        [
                            ConditionalContainer(
                                Window(
                                    width=1,
                                    char=border.MID_LEFT,
                                    style=self.add_style("class:left"),
                                ),
                                filter=border_left,
                            ),
                            DynamicContainer(lambda: self.body),
                            ConditionalContainer(
                                Window(
                                    width=1,
                                    char=border.MID_RIGHT,
                                    style=self.add_style("class:right"),
                                ),
                                filter=border_right,
                            )
                            # Padding is required to make sure that if the content is
                            # too small, the right frame border is still aligned.
                        ],
                        padding=0,
                    ),
                    ConditionalContainer(
                        VSplit(
                            [
                                ConditionalContainer(
                                    Window(
                                        width=1,
                                        height=1,
                                        char=border.BOTTOM_LEFT,
                                        style=self.add_style("class:left,bottom"),
                                    ),
                                    filter=border_bottom & border_left,
                                ),
                                ConditionalContainer(
                                    Window(
                                        char=border.BOTTOM_MID,
                                        style=self.add_style("class:bottom"),
                                    ),
                                    filter=border_bottom,
                                ),
                                ConditionalContainer(
                                    Window(
                                        width=1,
                                        height=1,
                                        char=border.BOTTOM_RIGHT,
                                        style=self.add_style("class:right,bottom"),
                                    ),
                                    filter=border_bottom & border_right,
                                ),
                            ],
                            # specifying height here will increase the rendering speed.
                            height=1,
                        ),
                        filter=border_bottom,
                    ),
                ],
            )
        else:
            self.container = body

    def add_style(self, extra: str) -> Callable[[], str]:
        """Return a function which adds a style string to the border style."""

        def _style() -> str:
            if callable(self.style):
                return f"{self.style()} {extra}"
            else:
                return f"{self.style} {extra}"

        return _style

    def __pt_container__(self) -> AnyContainer:
        """Return the border widget's container."""
        return self.container


class FocusedStyle(Container):
    """Apply a style to child containers when focused or hovered."""

    def __init__(
        self,
        body: AnyContainer,
        style_focus: str | Callable[[], str] = "class:focused",
        style_hover: str | Callable[[], str] = "",
    ) -> None:
        """Create a new instance of the widget.

        Args:
            body: The container to act on
            style_focus: The style to apply when the body has focus
            style_hover: The style to apply when the body is hovered
        """
        self.body = body
        self.style_focus = style_focus
        self.style_hover = style_hover
        self.hover = False
        self.has_focus = has_focus(self.body)

    def reset(self) -> None:
        """Reet the wrapped container."""
        to_container(self.body).reset()

    def preferred_width(self, max_available_width: int) -> Dimension:
        """Return the wrapped container's preferred width."""
        return to_container(self.body).preferred_width(max_available_width)

    def preferred_height(self, width: int, max_available_height: int) -> Dimension:
        """Return the wrapped container's preferred height."""
        return to_container(self.body).preferred_height(width, max_available_height)

    def write_to_screen(
        self,
        screen: Screen,
        mouse_handlers: MouseHandlers,
        write_position: WritePosition,
        parent_style: str,
        erase_bg: bool,
        z_index: int | None,
    ) -> None:
        """Draw the wrapped container with the additional style."""
        to_container(self.body).write_to_screen(
            screen,
            mouse_handlers,
            write_position,
            f"{parent_style} {self.get_style()}",
            erase_bg,
            z_index,
        )

        if self.style_hover:
            x_min = write_position.xpos
            x_max = x_min + write_position.width
            y_min = write_position.ypos
            y_max = y_min + write_position.height

            # Wrap mouse handlers to add "hover" class on hover
            def _wrap_mouse_handler(handler: Callable) -> MouseHandler:
                def wrapped_mouse_handler(
                    mouse_event: MouseEvent,
                ) -> NotImplementedOrNone:
                    result = handler(mouse_event)

                    if mouse_event.event_type == MouseEventType.MOUSE_MOVE:
                        app = get_app()
                        if (
                            mouse_event.position.x,
                            mouse_event.position.y,
                        ) == app.mouse_position:
                            app.mouse_limits = write_position
                            self.hover = True
                        else:
                            app.mouse_limits = None
                            self.hover = False
                        result = None
                    return result

                return wrapped_mouse_handler

            mouse_handler_wrappers: FastDictCache[
                tuple[Callable], MouseHandler
            ] = FastDictCache(get_value=_wrap_mouse_handler)
            for y in range(y_min, y_max):
                row = mouse_handlers.mouse_handlers[y]
                for x in range(x_min, x_max):
                    row[x] = mouse_handler_wrappers[(row[x],)]

    def get_style(self) -> str:
        """Determine the style to apply depending on the focus status."""
        style = ""
        if self.has_focus():
            style += (
                self.style_focus() if callable(self.style_focus) else self.style_focus
            )
        if self.hover:
            style += " " + (
                self.style_hover() if callable(self.style_hover) else self.style_hover
            )
        return style

    def get_children(self) -> list[Container]:
        """Return the list of child :class:`.Container` objects."""
        return [to_container(self.body)]


class DropShadow(Container):
    """A transparent container which makes the background darker."""

    def __init__(self, amount: float = 0.5) -> None:
        """Create a new instance."""
        self.amount = amount
        app = get_app()
        self.cp = app.color_palette
        self.renderer = app.renderer

    def reset(self) -> None:
        """Reet the wrapped container - here, do nothing."""

    def preferred_width(self, max_available_width: int) -> Dimension:
        """Return the wrapped container's preferred width."""
        return Dimension(weight=1)

    def preferred_height(self, width: int, max_available_height: int) -> Dimension:
        """Return the wrapped container's preferred height."""
        return Dimension(weight=1)

    def write_to_screen(
        self,
        screen: Screen,
        mouse_handlers: MouseHandlers,
        write_position: WritePosition,
        parent_style: str,
        erase_bg: bool,
        z_index: int | None,
    ) -> None:
        """Draw the wrapped container with the additional style."""
        attr_cache = self.renderer._attrs_for_style
        if attr_cache is not None:
            ypos = write_position.ypos
            xpos = write_position.xpos
            amount = self.amount
            for y in range(ypos, ypos + write_position.height):
                row = screen.data_buffer[y]
                for x in range(xpos, xpos + write_position.width):
                    char = row[x]
                    style = char.style
                    attrs = attr_cache[style]

                    if not (fg := attrs.color) or fg == "default":
                        color = self.cp.fg
                        style += f" fg:{color.darker(amount)}"
                    else:
                        try:
                            color = ColorPaletteColor(fg)
                        except ValueError:
                            pass
                        else:
                            style += f" fg:{color.darker(amount)}"

                    if not (bg := attrs.bgcolor) or bg == "default":
                        color = self.cp.bg
                        style += f" bg:{color.darker(amount)}"
                    else:
                        try:
                            color = ColorPaletteColor(bg)
                        except ValueError:
                            pass
                        else:
                            style += f" bg:{color.darker(amount)}"

                    row[x] = Char(char=char.char, style=style)

    def get_children(self) -> list[Container]:
        """Return an empty list of child :class:`.Container` objects."""
        return []


class Shadow:
    """Draw a shadow underneath/behind this container.

    This is a globally configurable version of the
    :py:class:`prompt_toolkit.widows.base.Shadow` class.
    """

    def __init__(self, body: AnyContainer) -> None:
        """Initialize a new drop-shadow container.

        Args:
            body: Another container object.
        """
        filter_ = get_app().config.filter("show_shadows")
        shadow = FloatContainer(
            content=body,
            floats=[
                Float(
                    bottom=-1,
                    height=1,
                    left=1,
                    right=0,
                    transparent=True,
                    content=DropShadow(),
                ),
                Float(
                    bottom=-1,
                    top=1,
                    width=1,
                    right=-1,
                    transparent=True,
                    content=DropShadow(),
                ),
            ],
        )

        def get_contents() -> AnyContainer:
            if filter_():
                return shadow
            else:
                return body

        self.container = DynamicContainer(get_contents)

    def __pt_container__(self) -> AnyContainer:
        """Return the container's content."""
        return self.container

    # ################################### Settings ####################################

    add_setting(
        name="show_shadows",
        flags=["--show-shadows"],
        type_=bool,
        help_="Show or hide shadows under menus and dialogs",
        default=True,
        description="""
            Sets whether shadows are shown under dialogs and popup-menus.
        """,
    )



"""Contain the browser CSS for formatting markdown."""

from __future__ import annotations

from functools import cache
from typing import TYPE_CHECKING

from prompt_toolkit.filters.utils import _always

from .html import CssSelector

if TYPE_CHECKING:
    from .html import CssSelectors


_MARKDOWN_CSS = {
    _always: {
        (
            (CssSelector(item="h1"),),
            (CssSelector(item="h2"),),
            (CssSelector(item="h3"),),
            (CssSelector(item="h4"),),
            (CssSelector(item="h5"),),
            (CssSelector(item="h6"),),
        ): {
            "text_align": "center",
        },
        ((CssSelector(item="h1"),),): {
            "font_weight": "bold",
            "text_decoration": "underline",
            "border_top_style": "double",
            "border_right_style": "double",
            "border_bottom_style": "double",
            "border_left_style": "double",
            "border_top_width": "0.34em",
            "border_right_width": "0.34em",
            "border_bottom_width": "0.34em",
            "border_left_width": "0.34em",
            "border_top_color": "ansiyellow",
            "border_right_color": "ansiyellow",
            "border_bottom_color": "ansiyellow",
            "border_left_color": "ansiyellow",
            "padding_bottom": "0",
        },
        ((CssSelector(item="h2"),),): {
            "font_weight": "bold",
            "border_top_style": "solid",
            "border_right_style": "solid",
            "border_bottom_style": "solid",
            "border_left_style": "solid",
            "border_top_width": "0.34em",
            "border_right_width": "0.34em",
            "border_bottom_width": "0.34em",
            "border_left_width": "0.34em",
            "border_top_color": "#888888",
            "border_right_color": "#888888",
            "border_bottom_color": "#888888",
            "border_left_color": "#888888",
            "padding_bottom": "0",
        },
        ((CssSelector(item="h3"),),): {
            "font_weight": "bold",
            "font_style": "italic",
            "border_top_style": "solid",
            "border_right_style": "solid",
            "border_bottom_style": "solid",
            "border_left_style": "solid",
            "border_top_width": "0.1em",
            "border_right_width": "0.1em",
            "border_bottom_width": "0.1em",
            "border_left_width": "0.1em",
            "border_top_color": "#888888",
            "border_right_color": "#888888",
            "border_bottom_color": "#888888",
            "border_left_color": "#888888",
            "margin_left": "auto",
            "margin_right": "auto",
            "padding_top": "0",
            "padding_right": "1em",
            "padding_bottom": "0",
            "padding_left": "1em",
        },
        ((CssSelector(item="h4"),),): {
            "text_weight": "bold",
            "text_decoration": "underline",
            "border_bottom_color": "#888888",
            "margin_top": "1rem",
            "margin_bottom": "1rem",
        },
        ((CssSelector(item="h5"),),): {
            "font_weight": "bold",
            "border_bottom_color": "#888888",
            "margin_top": "1rem",
            "margin_bottom": "1rem",
        },
        ((CssSelector(item="h6"),),): {
            "font_style": "italic",
            "border_bottom_color": "#888888",
            "margin_top": "1rem",
            "margin_bottom": "1rem",
        },
        (
            (
                CssSelector(item="ol"),
                CssSelector(item="li"),
                CssSelector(item="::marker"),
            ),
        ): {"color": "ansicyan"},
        (
            (
                CssSelector(item="ul"),
                CssSelector(item="li"),
                CssSelector(item="::marker"),
            ),
        ): {"color": "ansiyellow"},
        ((CssSelector(item="blockquote"),),): {
            "margin_top": "1em",
            "margin_bottom": "1em",
            "padding_left": "1em",
            "border_left_width": "thick",
            "border_left_style": "solid",
            "border_left_color": "darkmagenta",
        },
        ((CssSelector(item=".block"),),): {"display": "block"},
        ((CssSelector(item="td"),),): {
            "display": "table-cell",
            "border_top_style": "solid",
            "border_right_style": "solid",
            "border_bottom_style": "solid",
            "border_left_style": "solid",
            "border_top_width": "0.1em",
            "border_right_width": "0.1em",
            "border_bottom_width": "0.1em",
            "border_left_width": "0.1em",
            "padding_left": "1em",
            "padding_right": "1em",
            # "_pt_class": "markdown,table,border",
        },
        ((CssSelector(item="th"),),): {
            "display": "table-cell",
            "border_top_style": "solid",
            "border_right_style": "solid",
            "border_bottom_style": "solid",
            "border_left_style": "solid",
            "border_top_width": "0.34em",
            "border_right_width": "0.34em",
            "border_bottom_width": "0.34em",
            "border_left_width": "0.34em",
            "padding_left": "1em",
            "padding_right": "1em",
            "font_weight": "bold",
            # "_pt_class": "markdown,table,border",
        },
        ((CssSelector(item="code"),),): {
            "display": "inline",
            "_pt_class": "markdown,code",
        },
        ((CssSelector(item=".math"),),): {"text_transform": "latex"},
        ((CssSelector(item=".math.block"),),): {
            "text_align": "center",
        },
        (
            (
                CssSelector(item="pre"),
                CssSelector(comb=">", item="code"),
            ),
        ): {
            "display": "block",
            "border_top_style": "solid",
            "border_top_width": "1px",
            "border_right_style": "solid",
            "border_right_width": "1px",
            "border_bottom_style": "solid",
            "border_bottom_width": "1px",
            "border_left_style": "solid",
            "border_left_width": "1px",
            "_pt_class": "markdown,code,block",
        },
        (
            (
                CssSelector(item="pre"),
                CssSelector(comb=">", item="code"),
                CssSelector(item="*"),
            ),
        ): {
            "pt_class": "markdown,code,block",
        },
        ((CssSelector(item="img"),), (CssSelector(item="svg"),)): {
            "display": "inline-block",
            "overflow_x": "hidden",
            "overflow_y": "hidden",
            "vertical_align": "middle",
        },
    }
}


@cache
def get_markdown_file_css() -> CssSelectors:
    """Apply margins to the root - used if rendering a full markdown file."""
    return {
        _always: {
            **_MARKDOWN_CSS[_always],
            ((CssSelector(item="::root"),),): {
                "max_width": "100em",
                "margin_left": "auto",
                "margin_right": "auto",
            },
        }
    }
